from enum import Enum


class UpdateType(str, Enum):
    """
    This object represents the complete list of allowed update types
    """
    COMPETENCE = "competence"
    LINE = "line"
    SUBSCRIBER = "subscriber"
    SUBSCRIPTION = "subscription"
    SUPPORT_LINE = "support_line"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from .base import ConnectMethod
from ..types import UUID, InputFile


class SendFileCollegue(ConnectMethod[bool]):
    """
    Типы запросов: POST
    Описание: Отправка файла
    Название: SendFileCollegue
    URL: /v1/colleague/send/file/

    Метод позволяет отправить файл в чат сотруднику от имени другого сотрудника.
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1479442525/4.3.1.9.
    """

    __returning__ = bool

    recepient_id: UUID
    """ID получателя"""
    author_id: UUID
    """ID автора (от чьего имени отправляется файл)"""
    file_name: str
    """Имя файла"""
    comment: Optional[str] = None
    "Комментарий к файлу"

    file: Union[InputFile, str]
    """Файл в виде:
        URLInputFile(url="{url_ссылка_на_файл}")
        FSInputFile(path="{./путь_к_файлу}")
        BufferedInputFile(file=b"{содержимое}, filename="{название_файла}")
    """

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            recepient_id: UUID,
            author_id: UUID,
            file_name: str,
            comment: Optional[str] = None,
            file: Union[InputFile, str],
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                recepient_id=recepient_id,
                author_id=author_id,
                file_name=file_name,
                comment=comment,
                file=file,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from .base import ConnectMethod

from ..types import Lines


class GetLines(ConnectMethod[Lines]):
    """
    Типы запросов: GET
    Описание: Получение доступных линий поддержки
    Название: GetLines
    URL: /v1/line/

    Метод возвращает список линий поддержки.
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2156429313/4.2.3.5.
    """

    __returning__ = Lines



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .base import ConnectMethod

from ..types import UUID


class AppointStart(ConnectMethod[bool]):
    """
    Типы запросов: POST
    Описание: Назначить любого свободного специалиста
    Название: AppointStart
    URL: /v1/line/appoint/start/

    Метод Метод позволяет назначить любого свободного специалиста.
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1348632607/4.3.1.1.
    """

    __returning__ = bool

    line_id: UUID
    """ID линии поддержки"""
    user_id: UUID
    """ID пользователя"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            line_id: UUID,
            user_id: UUID,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                line_id=line_id,
                user_id=user_id,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from functools import partial
from typing import Any, Callable, List, Optional, Sequence, Union, cast, overload

from .base import (
    NextRequestMiddlewareType,
    RequestMiddlewareType,
)
from ....methods.base import ConnectType


class RequestMiddlewareManager(Sequence[RequestMiddlewareType]):
    def __init__(self) -> None:
        self._middlewares: List[RequestMiddlewareType] = []

    def register(
        self,
        middleware: RequestMiddlewareType,
    ) -> RequestMiddlewareType:
        self._middlewares.append(middleware)
        return middleware

    def unregister(self, middleware: RequestMiddlewareType) -> None:
        self._middlewares.remove(middleware)

    def __call__(
        self,
        middleware: Optional[RequestMiddlewareType] = None,
    ) -> Union[Callable[[RequestMiddlewareType], RequestMiddlewareType], RequestMiddlewareType]:
        if middleware is None:
            return self.register
        return self.register(middleware)

    @overload
    def __getitem__(self, item: int) -> RequestMiddlewareType:
        pass

    @overload
    def __getitem__(self, item: slice) -> Sequence[RequestMiddlewareType]:
        pass

    def __getitem__(
        self, item: Union[int, slice]
    ) -> Union[RequestMiddlewareType, Sequence[RequestMiddlewareType]]:
        return self._middlewares[item]

    def __len__(self) -> int:
        return len(self._middlewares)

    def wrap_middlewares(
        self,
        callback: NextRequestMiddlewareType[ConnectType],
        **kwargs: Any,
    ) -> NextRequestMiddlewareType[ConnectType]:
        middleware = partial(callback, **kwargs)
        for m in reversed(self._middlewares):
            middleware = partial(m, middleware)
        return cast(NextRequestMiddlewareType[ConnectType], middleware)



import logging

dispatcher = logging.getLogger("aio_connect.dispatcher")
event = logging.getLogger("aio_connect.event")
middlewares = logging.getLogger("aio_connect.middlewares")
webhook = logging.getLogger("aio_connect.webhook")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from ..types import UUID


class PartnerNotification(ConnectObject):
    """
    Название объекта: PartnerNotification
    Описание объекта: Рассылка

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2459074561/PartnerNotification
    """

    id: UUID
    """ID рассылки"""
    theme: str
    """Тема"""
    text: str
    """
    Текст рассылки с параметрами замены:
        <Фамилия>
        <Имя>
        <Отчество>
        <ДатаРождения>
        <Логин>
        <Должность>
        <Телефон>
        <EMail>
        <НазваниеКомпании>
        <ИННКомпании>
        <РегионКомпании>
    """
    description_hint: Optional[str] = None
    """Текст на кнопке"""
    description_url: Optional[str] = None
    """Ссылка, открываемая по кнопке. Так же как и text могут быть параметры."""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            id: UUID,
            theme: str,
            text: str,
            description_hint: Optional[str] = None,
            description_url: Optional[str] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                id=id,
                theme=theme,
                text=text,
                description_hint=description_hint,
                description_url=description_url,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from ..types import UUID


class File(ConnectObject):
    """
    Название объекта: File
    Описание объекта: Файл

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289388485/File
    """

    file_id: UUID
    """ID файла"""
    file_path: str
    """Адрес для скачивания"""
    file_name: str
    """Имя файла"""
    file_size: int
    """Размер файла, байт"""
    comment: Optional[str] = None
    """Сопроводительное сообщение к файлу"""
    preview_link: Optional[str] = None
    """Ссылка на превью для переданного изображения"""
    preview_link_hi: Optional[str] = None
    """Ссылка на превью в высоком качестве для переданного изображения"""
    preview_hi_sizes: Optional[str] = None
    """Размер превью (ширинаxвысота) в высоком качестве для переданного изображения"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            file_id: UUID,
            file_path: str,
            file_name: str,
            file_size: int,
            comment: Optional[str] = None,
            preview_link: Optional[str] = None,
            preview_link_hi: Optional[str] = None,
            preview_hi_sizes: Optional[str] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                file_id=file_id,
                file_path=file_path,
                file_name=file_name,
                file_size=file_size,
                comment=comment,
                preview_link=preview_link,
                preview_link_hi=preview_link_hi,
                preview_hi_sizes=preview_hi_sizes,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject
from ..types import UUID


class Call(ConnectObject):
    """
    Название объекта: Call
    Описание объекта: Звонок (голосовой вызов)

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1298628889/Call
    """

    src: UUID
    """ID инициатора"""
    dst: Optional[UUID] = None
    """ID адресата"""
    new_dst: Optional[UUID] = None
    """ID нового адресата (при перенаправлении вызова)"""
    billsec: Optional[int] = None
    """Длительность разговора, секунды"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            src: UUID,
            dst: Optional[UUID] = None,
            new_dst: Optional[UUID] = None,
            billsec: Optional[int] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                src=src,
                dst=dst,
                new_dst=new_dst,
                billsec=billsec,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from .base import ConnectMethod

from ..types import TicketShort


class GetTicket(ConnectMethod[TicketShort]):
    """
    Типы запросов: GET
    Описание: Получение заявки Service Desk со вложенными объектами по id
    Название: GetTicket
    URL: /v1/ticket/{id}/

    id: UUID = ID заявки

    Метод возвращает заявку Service Desk со вложенными объектами по id
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2492923976/4.2.3.10.+Service+Desk+ID
    """

    __returning__ = TicketShort



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from datetime import datetime
from ..types import (
    UUID,
    ServiceKind,
    LineShort,
    TicketType,
    TicketChannel,
    TicketStatus,
    User,
    TicketAdditionalFieldValue
)


class TicketShort(ConnectObject):
    """
    Название объекта: TicketShort
    Описание объекта: Заявка Service Desk (краткие сведения)

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2492923959/TicketShort
    """

    id: UUID
    """ID Заявки"""
    transaction_id: UUID
    """ID Транзакции"""
    number: int
    """Номер заявки"""
    created_at: datetime
    """Время создания"""
    updated_at: datetime
    """Время изменения"""
    description: Optional[str] = None
    """Описание"""
    priority: str
    """Приоритет:
        LOW - низкий
        STANDARD - Стандартный
        HIGH - Высокий
    """
    duration: int
    """Длительность работы, сек (0 - не задано)"""
    result: Optional[str] = None
    """Описание решения"""
    summary: Optional[str] = None
    """Тема"""
    deadline: Optional[str] = None
    """Срок"""
    kind: ServiceKind
    """Вид работ"""
    line: LineShort
    """Линия поддержки"""
    type: TicketType
    """Тип заявки"""
    channel: TicketChannel
    """Канал связи"""
    status: TicketStatus
    """Статус заявоки"""
    initiator: User
    """Заказчик"""
    author: Optional[User] = None
    """Автор(отсутствует, если заявка создана в учетной системе)"""
    executor: Optional[User] = None
    """Исполнитель"""
    fields: Optional[TicketAdditionalFieldValue] = None
    """Значения дополнительных полей заявки"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            id: UUID,
            transaction_id: UUID,
            number: int,
            created_at: datetime,
            updated_at: datetime,
            description: Optional[str] = None,
            priority: str,
            duration: int,
            result: Optional[str] = None,
            summary: Optional[str] = None,
            deadline: Optional[str] = None,
            kind: ServiceKind,
            line: LineShort,
            type: TicketType,
            channel: TicketChannel,
            status: TicketStatus,
            initiator: User,
            author: Optional[User] = None,
            executor: Optional[User] = None,
            fields: Optional[TicketAdditionalFieldValue] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                id=id,
                transaction_id=transaction_id,
                number=number,
                created_at=created_at,
                updated_at=updated_at,
                description=description,
                priority=priority,
                duration=duration,
                result=result,
                summary=summary,
                deadline=deadline,
                kind=kind,
                line=line,
                type=type,
                channel=channel,
                status=status,
                initiator=initiator,
                author=author,
                executor=executor,
                fields=fields,
                **__pydantic_kwargs,
            )



from typing import List, Literal, Optional, Union

from .base import ConnectObject
from .input_file import BufferedInputFile, FSInputFile, InputFile, URLInputFile
from .bot_command import BotCommand
from .update import Update
from .uuid import UUID, is_valid_uuid
from .answering import Answering
from .hook_type import HookType
# 4.2.2.1. Объекты
from .competence import Competence, Competences
from .subscriptions import Subscriptions
from .user_service_line import UserServiceLine
from .user import User, Users
from .line import Line, LineShort, Lines
from .call import Call
from .file import File
from .rda import Rda
from .service_request import ServiceRequest
from .treatment import Treatment, Treatments
from .data import Data
from .partner_notification import PartnerNotification
from .ticket_channel import TicketChannel
from .ticket_status import TicketStatus
from .ticket_type import TicketType
from .service_kind import ServiceKind
from .ticket_additional_field_value import TicketAdditionalFieldValue
from .ticket_short import TicketShort
# 4.2.2.2. Структура событий
from .type_competence import TypeCompetence
from .type_line import TypeLine
from .type_subscriber import TypeSubscriber
from .type_subscription import TypeSubscription
from .type_support_line import TypeSupportLine
# 4.3.2. Структуры данных для ботов
from .button import Button

__all__ = (
    "ConnectObject",
    "BufferedInputFile", "FSInputFile", "InputFile", "URLInputFile",
    "BotCommand",
    "Update",
    "UUID", "is_valid_uuid",
    "Answering",
    "HookType",
    # 4.2.2.1. Объекты
    "Competence", "Competences",
    "Subscriptions",
    "UserServiceLine",
    "User", "Users",
    "Line", "LineShort", "Lines",
    "Call",
    "File",
    "Rda",
    "ServiceRequest",
    "Treatment", "Treatments",
    "Data",
    "PartnerNotification",
    "TicketChannel",
    "TicketStatus",
    "TicketType",
    "ServiceKind",
    "TicketAdditionalFieldValue",
    "TicketShort",
    # 4.2.2.2. Структура событий
    "TypeCompetence",
    "TypeLine",
    "TypeSubscriber",
    "TypeSubscription",
    "TypeSupportLine",
    # 4.3.2. Структуры данных для ботов
    "Button",
)


# Load typing forward refs for every ConnectObject
for _entity_name in __all__:
    _entity = globals()[_entity_name]
    if not hasattr(_entity, "model_rebuild"):
        continue
    _entity.model_rebuild(
        _types_namespace={
            "List": List,
            "Optional": Optional,
            "Union": Union,
            "Literal": Literal,
            **{k: v for k, v in globals().items() if k in __all__},
        }
    )

del _entity
del _entity_name



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .base import ConnectObject

from ..types import UUID


class TicketType(ConnectObject):
    """
    Название объекта: TicketType
    Описание объекта: Типы заявок Service Desk

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2492956693/TicketType
    """

    id: UUID
    """ID Типа заявок"""
    name: str
    """Название"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            id: UUID,
            name: str,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                id=id,
                name=name,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from ..types import UUID


class Line(ConnectObject):
    """
    Название объекта: Line
    Описание объекта: Линия поддержки

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289420955/Line
    """

    line_id: UUID
    """ID линии поддержки"""
    name: str
    """Название"""
    allow_bot: bool
    """Дано разрешение на использование цифрового меню"""
    hook_bot: Optional[str] = None
    """Адрес, на который выставлен hook с типом bot"""
    hook_line: Optional[str] = None
    """Адрес, на который выставлен hook с типом line"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            line_id: UUID,
            name: str,
            allow_bot: bool,
            hook_bot: Optional[str] = None,
            hook_line: Optional[str] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                line_id=line_id,
                name=name,
                allow_bot=allow_bot,
                hook_bot=hook_bot,
                hook_line=hook_line,
                **__pydantic_kwargs,
            )


class LineShort(ConnectObject):
    """
    Название объекта: LineShort
    Описание объекта: Линия поддержки (краткие сведения)

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2493218817/LineShort
    """

    id: UUID
    """ID линии поддержки"""
    name: str
    """Название"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            id: UUID,
            name: str,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                id=id,
                name=name,
                **__pydantic_kwargs,
            )


class Lines(Line):
    """
    Название объекта: Lines
    Описание объекта: Линии поддержки

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2156429325/Lines

    Данные: Такие же как у Line
    """



import asyncio
import contextvars
import inspect
import warnings
from dataclasses import dataclass, field
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

from magic_filter.magic import MagicFilter as OriginalMagicFilter

from ...filters.base import Filter
from ...handlers import BaseHandler
from ...utils.magic_filter import MagicFilter
from ...utils.warnings import Recommendation

CallbackType = Callable[..., Any]


@dataclass
class CallableObject:
    callback: CallbackType
    awaitable: bool = field(init=False)
    params: Set[str] = field(init=False)
    varkw: bool = field(init=False)

    def __post_init__(self) -> None:
        callback = inspect.unwrap(self.callback)
        self.awaitable = inspect.isawaitable(callback) or inspect.iscoroutinefunction(callback)
        spec = inspect.getfullargspec(callback)
        self.params = {*spec.args, *spec.kwonlyargs}
        self.varkw = spec.varkw is not None

    def _prepare_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if self.varkw:
            return kwargs

        return {k: kwargs[k] for k in self.params if k in kwargs}

    async def call(self, *args: Any, **kwargs: Any) -> Any:
        wrapped = partial(self.callback, *args, **self._prepare_kwargs(kwargs))
        if self.awaitable:
            return await wrapped()

        loop = asyncio.get_event_loop()
        context = contextvars.copy_context()
        wrapped = partial(context.run, wrapped)
        return await loop.run_in_executor(None, wrapped)


@dataclass
class FilterObject(CallableObject):
    magic: Optional[MagicFilter] = None

    def __post_init__(self) -> None:
        if isinstance(self.callback, OriginalMagicFilter):
            # MagicFilter instance is callable but generates
            # only "CallOperation" instead of applying the filter
            self.magic = self.callback
            self.callback = self.callback.resolve
            if not isinstance(self.magic, MagicFilter):
                warnings.warn(
                    category=Recommendation,
                    message="You are using F provided by magic_filter package directly, "
                    "but it lacks `.as_()` extension."
                    "\n Please change the import statement: from `from magic_filter import F` "
                    "to `from aio_connect import F` to silence this warning.",
                    stacklevel=6,
                )

        super(FilterObject, self).__post_init__()

        if isinstance(self.callback, Filter):
            self.awaitable = True


@dataclass
class HandlerObject(CallableObject):
    filters: Optional[List[FilterObject]] = None

    def __post_init__(self) -> None:
        super(HandlerObject, self).__post_init__()
        callback = inspect.unwrap(self.callback)
        if inspect.isclass(callback) and issubclass(callback, BaseHandler):
            self.awaitable = True

    async def check(self, *args: Any, **kwargs: Any) -> Tuple[bool, Dict[str, Any]]:
        if not self.filters:
            return True, kwargs
        for event_filter in self.filters:
            check = await event_filter.call(*args, **kwargs)
            if not check:
                return False, kwargs
            if isinstance(check, dict):
                kwargs.update(check)
        return True, kwargs



from enum import Enum


class ContentType(str, Enum):
    """
    This object represents a type of content in message
    """

    UNKNOWN = "unknown"
    ANY = "any"

    CALL = "call"  # Тип line
    RDA = "rda"  # Тип line
    SERVICE_REQUEST = "service_request"  # Тип line
    TREATMENT = "treatment"  # Тип line
    DATA = "data"  # Тип line

    COMPETENCE = "competence"  # Тип competence
    USER = "user"  # Тип subscriber
    SUBSCRIPTION = "subscription"  # Тип subscription
    LINE = "line"  # Тип support_line

    ACTION = "action"  # Типы competence subscriber subscription support_line

    # Message type
    TEXT = "text"  # 1 - Текстовое сообщение
    NOTIFY = "notify"  # 2 - Рассылка
    REPORT = "report"  # 3 -
    BOT = "bot"  # 4 -

    RECORD = "record"  # 16 -

    QUALITY_WORK = "quality_work"  # 17 -

    CALL_START_WITH_NSS = "call_start_with_nss"  # 20 - Начало звонка с открытием обращения
    CALL_START_WITHOUT_NSS = "call_start_without_nss"  # 21 - Начало звонка без открытия обращения
    CALL_END_GOOD = "call_end_good"  # 22 - Удачное завершение звонка
    CALL_BADCALL_CANCELED = "call_badcall_canceled"  # 23 - Неудачное завершение звонка. Вызов отменен инициатором
    CALL_BADCALL_EXCESS = "call_badcall_excess"  # 24 - Неудачное завершение звонка. Абонент не ответил
    CALL_BADCALL_NONSS = "call_badcall_nonss"  # 25 - Неудачное завершение звонка. Нет свободных специалистов
    CALL_BADCALL_REJECTED = "call_badcall_rejected"  # 26 - Неудачное завершение звонка. Вызов отклонен адресатом
    CALL_REROUTING = "call_rerouting"  # 27 - Перевод звонка на специалиста
    CALL_REROUTING_VENDOR = "call_rerouting_vendor"  # 28 - Перевод звонка в компанию вендора
    CALL_REROUTING_COLL2COLL = "call_rerouting_coll2coll"  # 29 -
    CALL_REROUTING_WITHOUTEND = "call_rerouting_withoutend"  # 30 - Перевод звонка без завершения(когда абонент не взял
    # трубку, а сразу перевел звонок)
    CALL_REROUTING_VENDORFRAN_WE = "call_rerouting_vendorfran_we"  # 31 - Перевод звонка без завершения на вендора(когда
    # абонент не взял трубку, а сразу перевел звонок)

    CALL_TECHNICAL_PROBLEM = "call_technical_problem"  # 32 - Нет соединения с голосовым сервером
    CALL_TECHNICAL_PROBLEM_33 = "call_technical_problem_33"  # 33 -
    CALL_TECHNICAL_PROBLEM_34 = "call_technical_problem_34"  # 34 -
    CALL_TECHNICAL_PROBLEM_35 = "call_technical_problem_35"  # 35 -
    CALL_TECHNICAL_PROBLEM_NO_AUDIO = "call_technical_problem_no_audio"  # 36 - Неудачное завершение. У абонента нет
    # аудио устройства
    CALL_TECHNICAL_PROBLEM_37 = "call_technical_problem_37"  # 37 -
    LINE_CALLUNAVAIL = "line_callunavail"  # 38 - Недоступная линия по звонку (нерабочее время)

    CONFERENCE_STARTCALL = "conference_startcall"  # 40 -
    CONFERENCE_ENDCALL = "conference_endcall"  # 41 -
    CONFERENCE_BADCALL_CANCELED = "conference_badcall_canceled"  # 42 -
    CONFERENCE_BADCALL_EXCESS = "conference_badcall_excess"  # 43 -
    CONFERENCE_BADCALL_TECHNICALPROBLEM = "conference_badcall_technicalproblem"  # 44 -

    RDA_START_WITH_NSS = "rda_start_with_nss"  # 50 - Начало сеанса удаленного доступа с сервисным специалистом
    RDA_START_WITHOUT_NSS = "rda_start_without_nss"  # 51 - Начало сеанса удаленного доступа без сервисного специалиста
    # (Специалист принудительно назначается пользователю сеансом удаленного доступа)
    RDA_END_WITHT_RANSFER_FILES = "rda_end_witht_ransfer_files"  # 52 - Окончание сеанса удаленного доступа с передачей
    # файлов
    RDA_END_WITHTOUT_RANSFER_FILES = "rda_end_withtout_ransfer_files"  # 53 - Окончание сеанса удаленного доступа без
    # передачей файлов
    RDA_BAD_CANCELED = "rda_bad_canceled"  # 54 - Неудачное завершение удаленного доступа. Сеанс отменен инициатором
    RDA_BAD_REJECTED = "rda_bad_rejected"  # 55 - Неудачное завершение удаленного доступа. Сеанс отклонен принимающим
    RDA_BAD_EXCESS = "rda_bad_excess"  # 56 - Неудачное завершение удаленного доступа. Сеанс не состоялся по таймауту
    RDA_BAD = "rda_bad"  # 57 - Неудачное завершение удаленного доступа
    RDA_P2P = "rda_p2p"  # 58 -
    RDA_BAD_OLD_SERVICE = "rda_bad_old_service"  # 59 - Неудачное завершение - устаревшая версия службы
    RDA_BAD_NO_SERVICE = "rda_bad_no_service"  # 60 - Неудачное завершение - служба не установлена
    RDA_BAD_OLD_COMPONENT = "rda_bad_old_component"  # 61 - Неудачное завершение - устаревшие версии компонентов
    RDA_BAD_NO_FILES = "rda_bad_no_files"  # 62 - Неудачное завершение - отсутствуют файлы компонентов УД

    TRANSFERFILES = "transferfiles"  # 70 - Передача файла через чат

    LINE_USERINIT = "line_userinit"  # 80 - Назначение специалиста системой. Инициатором был пользователь
    LINE_SPECINIT = "line_specinit"  # 81 - Специалист назначился. Инициатором был специалист
    LINE_SPECDEL = "line_specdel"  # 82 - Завершение работы специалиста (Закрытие обращения)
    LINE_NONSS = "line_nonss"  # 83 - Обращение поступило в очередь. Нет свободных специалистов

    LINE_REROUTINGSPEC = "line_reroutingspec"  # 84 - Перевод обращения на специалиста

    LINE_REROUTING_VENDOR = "line_rerouting_vendor"  # 85 - Перевод обращения в компанию вендора

    LINE_SPECFOUND = "line_specfound"  # 86 - Для обращения в очереди появился свободный специалист
    LINE_CHATUNAVAIL = "line_chatunavail"  # 87 - Недоступность линии (нерабочее время)
    LINE_REROUTE_UNAVAIL = "line_reroute_unavail"  # 88 - Недоступность линии по переводу. При переводе обращения в
    # компанию вендора попали в нерабочее время

    LINE_REROUTING_OTHERSERVICE = "line_rerouting_otherservice"  # 89 - Перевод в другую линию поддержки

    LINE_CLOSED_NO_ACTIVITY = "line_closed_no_activity"  # 90 - Обращение закрыто автоматически по отсутствию активности
    # в чате
    LINE_CLOSED_REMOVE_SERVICE = "line_closed_remove_service"  # 91 - Обращение закрыто автоматически, т.к.удалена линия
    # поддержки
    LINE_CLOSED_REMOVE_SUBSCRIPTION = "line_closed_remove_subscription"  # 92 - Обращение закрыто автоматически, т.к.
    # удалена подписка пользователя
    LINE_CLOSED_REMOVE_USER = "line_closed_remove_user"  # 93 - Обращение закрыто автоматически, т.к.удален пользователь

    BUSSINES_CONTACT_ADD = "bussines_contact_add"  # 100 -
    BUSSINES_CONTACT_CANCEL = "bussines_contact_cancel"  # 101 -
    BUSSINES_CONTACT_HIDE = "bussines_contact_hide"  # 102 -
    BUSSINES_CONTACT_ACCEPT = "bussines_contact_accept"  # 103 -
    BUSSINES_CONTACT_REMOVE = "bussines_contact_remove"  # 104 -

    CONFERENCE_NEW = "conference_new"  # 110 -
    CONFERENCE_CHANGENAME = "conference_changename"  # 111 -
    CONFERENCE_NEWAVATAR = "conference_newavatar"  # 112 -
    CONFERENCE_ADD = "conference_add"  # 113 -
    CONFERENCE_DELETE = "conference_delete"  # 114 -
    CONFERENCE_LEAVE = "conference_leave"  # 115 -
    CONFERENCE_CLOSE = "conference_close"  # 116 -
    CONFERENCE_SETADMIN = "conference_setadmin"  # 117 -
    CONFERENCE_CALLCANCEL_BY_MEMBERS = "conference_callcancel_by_members"  # 118 -
    CONFERENCE_ADD_FROM_LINK = "conference_add_from_link"  # 119 -

    SERVICE_REQUEST_ADD = "service_request_add"  # 121 - Создание заявки Service Desk
    # 122 - Изменение заявки Service Desk
    # 123 - Завершение заявки Service Desk
    # 124 - Отмена заявки Service Desk

    LINE_REROUTING_TO_BOT = "line_rerouting_to_bot"  # 200 - Перевод обращения специалистом на бота



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from datetime import datetime
from ..types import UUID


class UserServiceLine(ConnectObject):
    """
    Название объекта: UserServiceLine
    Описание объекта: Линии, подключенные пользователям

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289322629/UserServiceLine
    """

    service_id: UUID
    """ID линии поддержки"""
    service_user_id: UUID
    """ID пользователя"""
    activate_date: datetime
    """Дата начала подписки"""
    expires_date: Optional[datetime] = None
    """Дата окончания подписки, null когда бессрочная подписка"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            service_id: UUID,
            service_user_id: UUID,
            activate_date: datetime,
            expires_date: Optional[datetime] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                service_id=service_id,
                service_user_id=service_user_id,
                activate_date=activate_date,
                expires_date=expires_date,
                **__pydantic_kwargs,
            )



from __future__ import annotations

import re
from dataclasses import dataclass, field, replace
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Match,
    Optional,
    Pattern,
    Sequence,
    Union,
    cast,
)

from magic_filter import MagicFilter

from .base import Filter
from ..types import BotCommand, TypeLine

if TYPE_CHECKING:
    from .. import Bot

CommandPatternType = Union[str, re.Pattern, BotCommand]


class CommandException(Exception):
    pass


class Command(Filter):
    """
    This filter can be helpful for handling commands from the text messages.
    """

    __slots__ = (
        "commands",
        "prefix",
        "ignore_case",
        "magic",
    )

    def __init__(
        self,
        *values: CommandPatternType,
        commands: Optional[Union[Sequence[CommandPatternType], CommandPatternType]] = None,
        prefix: str = "/",
        ignore_case: bool = False,
        magic: Optional[MagicFilter] = None,
    ):
        """
        List of commands (string or compiled regexp patterns)

        :param prefix: Prefix for command.
            Prefix is always a single char but here you can pass all of allowed prefixes,
            for example: :code:`"/!"` will work with commands prefixed
            by :code:`"/"` or :code:`"!"`.
        :param ignore_case: Ignore case (Does not work with regexp, use flags instead)
        :param magic: Validate command object via Magic filter after all checks done
        """
        if commands is None:
            commands = []
        if isinstance(commands, (str, re.Pattern, BotCommand)):
            commands = [commands]

        if not isinstance(commands, Iterable):
            raise ValueError(
                "Command filter only supports str, re.Pattern, BotCommand object"
                " or their Iterable"
            )

        items = []
        for command in (*values, *commands):
            if isinstance(command, BotCommand):
                command = command.command
            if not isinstance(command, (str, re.Pattern)):
                raise ValueError(
                    "Command filter only supports str, re.Pattern, BotCommand object"
                    " or their Iterable"
                )
            if ignore_case and isinstance(command, str):
                command = command.casefold()
            items.append(command)

        if not items:
            raise ValueError("At least one command should be specified")

        self.commands = tuple(items)
        self.prefix = prefix
        self.ignore_case = ignore_case
        self.magic = magic

    def __str__(self) -> str:
        return self._signature_to_string(
            *self.commands,
            prefix=self.prefix,
            ignore_case=self.ignore_case,
            magic=self.magic,
        )

    async def __call__(self, line: TypeLine, bot: Bot) -> Union[bool, Dict[str, Any]]:
        if not isinstance(line, TypeLine):
            return False

        text = line.text
        if not text:
            return False

        try:
            command = await self.parse_command(text=text)
        except CommandException:
            return False
        result = {"command": command}
        if command.magic_result and isinstance(command.magic_result, dict):
            result.update(command.magic_result)
        return result

    def extract_command(self, text: str) -> CommandObject:
        # First step: separate command with arguments
        # "/command arg1 arg2" -> "/command", ["arg1 arg2"]
        try:
            full_command, *args = text.split(maxsplit=1)
        except ValueError:
            raise CommandException("not enough values to unpack")

        # Separate command into valuable parts
        # "/command" -> "/", ("command")
        prefix, (command) = full_command[0], full_command[1:]
        return CommandObject(
            prefix=prefix,
            command=command,
            args=args[0] if args else None,
        )

    def validate_prefix(self, command: CommandObject) -> None:
        if command.prefix not in self.prefix:
            raise CommandException("Invalid command prefix")

    def validate_command(self, command: CommandObject) -> CommandObject:
        for allowed_command in cast(Sequence[CommandPatternType], self.commands):
            # Command can be presented as regexp pattern or raw string
            # then need to validate that in different ways
            if isinstance(allowed_command, Pattern):  # Regexp
                result = allowed_command.match(command.command)
                if result:
                    return replace(command, regexp_match=result)

            command_name = command.command
            if self.ignore_case:
                command_name = command_name.casefold()

            if command_name == allowed_command:  # String
                return command
        raise CommandException("Command did not match pattern")

    async def parse_command(self, text: str) -> CommandObject:
        """
        Extract command from the text and validate

        :param text:
        :return:
        """
        command = self.extract_command(text)
        self.validate_prefix(command=command)
        command = self.validate_command(command)
        command = self.do_magic(command=command)
        return command  # noqa: RET504

    def do_magic(self, command: CommandObject) -> Any:
        if self.magic is None:
            return command
        result = self.magic.resolve(command)
        if not result:
            raise CommandException("Rejected via magic filter")
        return replace(command, magic_result=result)


@dataclass(frozen=True)
class CommandObject:
    """
    Instance of this object is always has command and it prefix.
    Can be passed as keyword argument **command** to the handler
    """

    prefix: str = "/"
    """Command prefix"""
    command: str = ""
    """Command without prefix"""
    args: Optional[str] = field(repr=False, default=None)
    """Command argument"""
    regexp_match: Optional[Match[str]] = field(repr=False, default=None)
    """Will be presented match result if the command is presented as regexp in filter"""
    magic_result: Optional[Any] = field(repr=False, default=None)

    @property
    def text(self) -> str:
        """
        Generate original text from object
        """
        line = self.prefix + self.command
        if self.args:
            line += " " + self.args
        return line



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .base import ConnectObject

from ..types import UUID


class Competence(ConnectObject):
    """
    Название объекта: Competence
    Описание объекта: Компетенция

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289322615/Competence
    """

    line_id: UUID
    """ID линии поддержки"""
    specialist_id: UUID
    """ID пользователя"""
    pool_priority: int
    """
    Приоритет специалиста:
        0 - наблюдатель
        1 - высокий
        2 - средний
        3 - стандартный
    """
    is_franch_spec: bool
    """Является ли специалистом франа"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            line_id: UUID,
            specialist_id: UUID,
            pool_priority: int,
            is_franch_spec: bool,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                line_id=line_id,
                specialist_id=specialist_id,
                pool_priority=pool_priority,
                is_franch_spec=is_franch_spec,
                **__pydantic_kwargs,
            )


class Competences(Competence):
    """
    Название объекта: Competences
    Описание объекта: Компетенции

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2490695681/Competences

    Данные: Такие же как у Competence
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, cast

from ..utils.mypy_hacks import lru_cache
from .base import ConnectObject

if TYPE_CHECKING:
    from .type_competence import TypeCompetence
    from .type_line import TypeLine
    from .type_subscriber import TypeSubscriber
    from .type_subscription import TypeSubscription
    from .type_support_line import TypeSupportLine


class Update(ConnectObject):
    """
    This `object` represents an incoming update.

    At most **one** of the optional parameters can be present in any given update.
    """

    event_type: str
    """Тип события:
        line - событие по линии поддержки
        subscriber - изменения в профиле пользователя
        support_line - изменения линии поддержки
        competence - изменение компетенция
        subscription - изменения в подписках
    """
    event_source: str
    """Для подписки какого типа сгенерировано событие:
        line - событие на подписки с типом line
        bot - событие на подписки с типом bot
    """

    competence: Optional[TypeCompetence] = None
    line: Optional[TypeLine] = None
    subscriber: Optional[TypeSubscriber] = None
    subscription: Optional[TypeSubscription] = None
    support_line: Optional[TypeSupportLine] = None

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            event_type: str,
            event_source: str,

            competence: Optional[TypeCompetence] = None,
            line: Optional[TypeLine] = None,
            subscriber: Optional[TypeSubscriber] = None,
            subscription: Optional[TypeSubscription] = None,
            support_line: Optional[TypeSupportLine] = None,

            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                event_type=event_type,
                event_source=event_source,

                competence=competence,
                line=line,
                subscriber=subscriber,
                subscription=subscription,
                support_line=support_line,

                **__pydantic_kwargs,
            )

    @property
    @lru_cache()
    def get_event_type(self) -> str:
        """
        Detect update type
        If update type is unknown, raise UpdateTypeLookupError

        :return:
        """

        if self.competence:
            return "competence"
        if self.line:
            return "line"
        if self.subscriber:
            return "subscriber"
        if self.subscription:
            return "subscription"
        if self.support_line:
            return "support_line"

        raise UpdateTypeLookupError("Update does not contain any known event type.")

    @property
    def event(self) -> ConnectObject:
        return cast(ConnectObject, getattr(self, self.get_event_type))


class UpdateTypeLookupError(LookupError):
    """Update does not contain any known event type."""



from __future__ import annotations

from .base import ConnectMethod

from ..types import User


class GetSubscriber(ConnectMethod[User]):
    """
    Типы запросов: GET
    Описание: Получение информации о пользователе
    Название: GetSubscriber
    URL: /v1/line/subscriber/{user_id}/

    user_id: UUID = ID пользователя, являющегося получателем любой линии поддержки

    Метод возвращает информацию о пользователе
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289388228/4.2.3.2.
    """

    __returning__ = User



import uuid
from typing import TypeVar, Union
from pydantic import UUID3, UUID4, UUID5

UUID = TypeVar("UUID", bound=Union[UUID5, UUID4, UUID3, str])


def is_valid_uuid(uuid_to_test: UUID) -> UUID:
    if uuid_to_test:
        try:
            if uuid.UUID(str(uuid_to_test)).version:
                return str(uuid_to_test)
        except ValueError as e:
            raise ValueError(e, uuid_to_test)
        return -1



from __future__ import annotations

import asyncio
import warnings
from asyncio import Event, Lock
from typing import Any, Dict, Optional, Set

from .. import loggers
from ..client.bot import Bot
from ..exceptions import ConnectAPIError
from ..fsm.middleware import FSMContextMiddleware
from ..fsm.storage.base import BaseEventIsolation, BaseStorage
from ..fsm.storage.memory import DisabledEventIsolation, MemoryStorage
from ..methods import ConnectMethod
from ..types import Update
from ..types.update import UpdateTypeLookupError
from .event.bases import UNHANDLED, SkipHandler
from .event.connect import ConnectEventObserver
from .middlewares.error import ErrorsMiddleware
from .middlewares.user_context import UserContextMiddleware
from .router import Router


class Dispatcher(Router):
    """
    Root router
    """

    def __init__(
        self,
        *,  # * - Preventing to pass instance of Bot to the FSM storage
        storage: Optional[BaseStorage] = None,
        events_isolation: Optional[BaseEventIsolation] = None,
        disable_fsm: bool = False,
        name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Root router

        :param storage: Storage for FSM
        :param fsm_strategy: FSM strategy
        :param events_isolation: Events isolation
        :param disable_fsm: Disable FSM, note that if you disable FSM
            then you should not use storage and events isolation
        :param kwargs: Other arguments, will be passed as keyword arguments to handlers
        """
        super(Dispatcher, self).__init__(name=name)

        if storage and not isinstance(storage, BaseStorage):
            raise TypeError(
                f"FSM storage should be instance of 'BaseStorage' not {type(storage).__name__}"
            )

        # 1C-Connect API provides originally only one event type - Update
        # For making easily interactions with events here is registered handler which helps
        # to separate Update to different event types
        self.update = self.observers["update"] = ConnectEventObserver(
            router=self, event_name="update"
        )
        self.update.register(self._listen_update)

        # Error handlers should work is out of all other functions
        # and should be registered before all others middlewares
        self.update.outer_middleware(ErrorsMiddleware(self))

        # User context middleware makes small optimization for all other builtin
        # middlewares via caching the user and chat instances in the event context
        self.update.outer_middleware(UserContextMiddleware())

        # FSM middleware should always be registered after User context middleware
        # because here is used context from previous step
        self.fsm = FSMContextMiddleware(
            storage=storage or MemoryStorage(),
            events_isolation=events_isolation or DisabledEventIsolation(),
        )
        if not disable_fsm:
            # Note that when FSM middleware is disabled, the event isolation is also disabled
            # Because the isolation mechanism is a part of the FSM
            self.update.outer_middleware(self.fsm)
        self.shutdown.register(self.fsm.close)

        self.workflow_data: Dict[str, Any] = kwargs
        self._running_lock = Lock()
        self._stop_signal: Optional[Event] = None
        self._stopped_signal: Optional[Event] = None
        self._handle_update_tasks: Set[asyncio.Task[Any]] = set()

    def __getitem__(self, item: str) -> Any:
        return self.workflow_data[item]

    def __setitem__(self, key: str, value: Any) -> None:
        self.workflow_data[key] = value

    def __delitem__(self, key: str) -> None:
        del self.workflow_data[key]

    def get(self, key: str, /, default: Optional[Any] = None) -> Optional[Any]:
        return self.workflow_data.get(key, default)

    @property
    def storage(self) -> BaseStorage:
        return self.fsm.storage

    @property
    def parent_router(self) -> Optional[Router]:
        """
        Dispatcher has no parent router and can't be included to any other routers or dispatchers

        :return:
        """
        return None  # noqa: RET501

    @parent_router.setter
    def parent_router(self, value: Router) -> None:
        """
        Dispatcher is root Router then configuring parent router is not allowed

        :param value:
        :return:
        """
        raise RuntimeError("Dispatcher can not be attached to another Router.")

    async def feed_update(self, bot: Bot, update: Update, **kwargs: Any) -> Any:
        """
        Main entry point for incoming updates
        Response of this method can be used as Webhook response

        :param bot:
        :param update:
        """
        loop = asyncio.get_running_loop()
        handled = False
        start_time = loop.time()

        if update.bot != bot:
            # Re-mounting update to the current bot instance for making possible to
            # use it in shortcuts.
            # Here is update is re-created because we need to propagate context to
            # all nested objects and attributes of the Update, but it
            # is impossible without roundtrip to JSON :(
            # The preferred way is that pass already mounted Bot instance to this update
            # before call feed_update method
            update = Update.model_validate(update.model_dump(), context={"bot": bot})

        try:
            response = await self.update.wrap_outer_middleware(
                self.update.trigger,
                update,
                {
                    **self.workflow_data,
                    **kwargs,
                    "bot": bot,
                },
            )
            handled = response is not UNHANDLED
            return response
        finally:
            finish_time = loop.time()
            duration = (finish_time - start_time) * 1000
            loggers.event.info(
                "Update is %s. Duration %d ms",
                "handled" if handled else "not handled",
                duration
            )

    async def feed_raw_update(self, bot: Bot, update: Dict[str, Any], **kwargs: Any) -> Any:
        """
        Main entry point for incoming updates with automatic Dict->Update serializer

        :param bot:
        :param update:
        :param kwargs:
        """
        parsed_update = Update.model_validate(update, context={"bot": bot})
        return await self.feed_update(bot=bot, update=parsed_update, **kwargs)

    async def _listen_update(self, update: Update, **kwargs: Any) -> Any:
        """
        Main updates listener

        Workflow:
        - Detect content type and propagate to observers in current router
        - If no one filter is pass - propagate update to child routers as Update

        :param update:
        :param kwargs:
        :return:
        """
        try:
            update_type = update.event_type
            event = update.event
        except UpdateTypeLookupError as e:
            warnings.warn(
                "Detected unknown update type.\n"
                "Seems like Connect Bot API was updated and you have "
                "installed not latest version of aio_connect framework"
                f"\nUpdate: {update.model_dump_json(exclude_unset=True)}",
                RuntimeWarning,
            )
            raise SkipHandler() from e

        kwargs.update(event_update=update)

        return await self.propagate_event(update_type=update_type, event=event, **kwargs)

    @classmethod
    async def silent_call_request(cls, bot: Bot, result: ConnectMethod[Any]) -> None:
        """
        Simulate answer into WebHook

        :param bot:
        :param result:
        :return:
        """
        try:
            await bot(result)
        except ConnectAPIError as e:
            # In due to WebHook mechanism doesn't allow getting response for
            # requests called in answer to WebHook request.
            # Need to skip unsuccessful responses.
            # For debugging here is added logging.
            loggers.event.error("Failed to make answer: %s: %s", e.__class__.__name__, e)

    async def _feed_webhook_update(self, bot: Bot, update: Update, **kwargs: Any) -> Any:
        """
        The same with `Dispatcher.process_update()` but returns real response instead of bool
        """
        try:
            return await self.feed_update(bot, update, **kwargs)
        except Exception as e:
            loggers.event.exception(
                "Cause exception while process update \n%s: %s",
                e.__class__.__name__,
                e,
            )
            raise



from __future__ import annotations

import aiohttp

import pathlib
from contextlib import asynccontextmanager
from typing import (
    Any,
    AsyncIterator,
    List,
    Optional,
    TypeVar,
    Union,
    Dict,
    Literal
)

from ..methods import (
    ConnectMethod,
    # 4.2.1 Команды к механизму трансляции
    SetHook,
    DelAllHook,
    DelHook,
    # 4.2.3 Команды для уточнения информации
    GetTreatments,
    GetSubscriber,
    GetSubscribers,
    GetSubscriptions,
    GetLines,
    GetSpecialist,
    GetSpecialists,
    GetSpecialistsAvailable,
    GetCompetences,
    GetTicket,
    GetTicketByNumber,
    # 4.3.1. Команды внешних ботов
    AppointStart,
    AppointSpec,
    DropTreatment,
    SendMessageLine,
    SendFileLine,
    SendImageLine,
    DropKeyboard,
    SendMessageColleague,
    SendFileCollegue,
    SendImageColleague,
    SendMessageConference,
    SendFileConference,
    SendImageConference,
    QuestionAndAnswering, QuestionAndAnsweringSelected,
)
from ..types import (
    InputFile,
    UUID, is_valid_uuid,
    Answering,
    HookType,
    # 4.2.2.1. Объекты
    Competences,
    Subscriptions,
    User, Users,
    Lines,
    Treatments,
    TicketShort,
    # 4.3.2. Структуры данных для ботов
    Button,
)
from .session.aiohttp import AiohttpSession
from .session.base import BaseSession

T = TypeVar("T")


class Bot:
    def __init__(
        self,
        api_login: str,
        api_password: str,
        line_id: str,
        base: str,

        session: Optional[BaseSession] = None,
    ) -> None:
        """
        Bot class

        :param api_login:
        :param api_password:
        :param line_id:
        :param base: API server
        """

        if api_login and api_password:
            self.auth = aiohttp.BasicAuth(login=api_login, password=api_password)

        if session is None:
            session = AiohttpSession()

        self.api_login = api_login
        self.api_password = api_password
        self.line_id = line_id
        self.base = base
        self.session = session

    @asynccontextmanager
    async def context(self, auto_close: bool = True) -> AsyncIterator[Bot]:
        """
        Generate bot context

        :param auto_close: close session on exit
        :return:
        """
        try:
            yield self
        finally:
            if auto_close:
                await self.session.close()

    async def __call__(
        self,
        method: ConnectMethod[T],
        type_request: Literal["POST", "GET", "DELETE", "PUT", "POST-With-Attach"],
        path: str,
        request_timeout: Optional[int] = None
    ) -> T:
        """
        Call API method

        :param method:
        :param type_request: тип запроса
        :param path: URL после обращения к API
        :return:
        """
        return await self.session(self, method, timeout=request_timeout, type_request=type_request, path=path)

    async def download_file(
        self,
        file_path: str,
        destination: Union[pathlib.Path, str],
        timeout: int = 30,
    ) -> bool:
        """
        Download file by file_path to destination.

        :param file_path: Ссылка на скачивание
        :param destination: Путь и (или) имя файла
        :param timeout: Total timeout in seconds, defaults to 30
        """

        async with aiohttp.ClientSession() as session:
            async with session.get(
                    auth=self.auth, url=file_path, timeout=timeout if timeout is None else timeout
            ) as resp:
                if resp.status == 200:
                    content = await resp.read()
                    with open(destination, 'wb') as file:
                        file.write(content)
                    return True
                else:
                    print(f'Ошибка скачивания файла. Код статуса: {resp.status}')
        return False

    """
    4.2.1 Команды к механизму трансляции

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1349812242/4.2.1.
    """

    async def set_hook(
        self,
        url: str,
        type: HookType,
        id: Optional[UUID] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод устанавливает webhook для получения событий.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289388306/4.2.1.1.

        :param url: URL WebHook. На этот адрес будут прилетать все события POST-запросами.
        :param type: Тип подписки: bot (на события для чат бота для указанной линии поддержки (обязателен id)),
            line (все события по линии(ям) поддержки)
        :param id: ID объекта, на который осуществляется подписка: ID линии поддержки,
            если type == line и не указан id, то подписка производится на все линии, к которым имеется доступ
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SetHook(
            url=url,
            type=type,
            id=is_valid_uuid(id),
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/hook/")

    async def del_all_hook(
        self,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод удаляет адреса всех webhook (кроме bot).

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289355653/4.2.1.2.

        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = DelAllHook()
        return await self(call, request_timeout=request_timeout,
                          type_request="DELETE", path="/v1/hook/")

    async def del_hook(
        self,
        id: UUID,
        type: str,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод удаляет адрес конкретного webhook.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289355682/4.2.1.3.

        :param id: ID объекта
        :param type: Тип WebHook: bot (на события для чат бота), line (все события по линии(ям) поддержки)
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = DelHook()
        return await self(call, request_timeout=request_timeout,
                          type_request="DELETE", path=f"/v1/hook/{type}/{is_valid_uuid(id)}/")

    """
    4.2.3 Команды для уточнения информации

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1349877812/4.2.3.
    """

    async def get_treatments(
        self,
        line_id: Optional[UUID] = None,
        user_id: Optional[UUID] = None,
        request_timeout: Optional[int] = None,
    ) -> Treatments:
        """
        Метод возвращает список открытых обращений.
        Принимает два необязательных параметра line_id и user_id.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289355600/4.2.3.1.

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param request_timeout: Request timeout
        :return: Return :code:`Treatment`.
        """

        call = GetTreatments(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id)
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path="/v1/line/treatment/")

    async def get_subscriber(
        self,
        user_id: UUID,
        request_timeout: Optional[int] = None,
    ) -> User:
        """
        Метод возвращает информацию о пользователе

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289388228/4.2.3.2.

        :param user_id: ID пользователя, являющегося получателем любой линии поддержки
        :param request_timeout: Request timeout
        :return: Return :code:`User`.
        """

        call = GetSubscriber()
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path=f"/v1/line/subscriber/{is_valid_uuid(user_id)}/")

    async def get_subscribers(
        self,
        request_timeout: Optional[int] = None,
    ) -> Users:
        """
        Метод возвращает информацию о пользователях

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2068938753/4.2.3.3.

        :param request_timeout: Request timeout
        :return: Return :code:`Users`.
        """

        call = GetSubscribers()
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path="/v1/line/subscribers/")

    async def get_subscriptions(
        self,
        user_id: Optional[UUID] = None,
        client_id: Optional[UUID] = None,
        line_id: Optional[UUID] = None,
        request_timeout: Optional[int] = None,
    ) -> Subscriptions:
        """
        Метод возвращает информацию о получаемым линиям пользователями,
        с возможностью отфильтровать по пользователю, линии или клиенту.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2322923521/4.2.3.4.

        :param user_id: ID пользователя
        :param client_id: ID клиента
        :param line_id: ID линии поддержки
        :param request_timeout: Request timeout
        :return: Return :code:`Subscriptions`.
        """

        call = GetSubscriptions(
            user_id=is_valid_uuid(user_id),
            client_id=is_valid_uuid(client_id),
            line_id=is_valid_uuid(line_id)
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path="/v1/line/subscriptions/")

    async def get_lines(
        self,
        request_timeout: Optional[int] = None,
    ) -> Lines:
        """
        Метод возвращает список линий поддержки.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2156429313/4.2.3.5.

        :param request_timeout: Request timeout
        :return: Returns :code:`Lines`.
        """

        call = GetLines()
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path="/v1/line/")

    async def get_specialist(
        self,
        user_id: UUID,
        request_timeout: Optional[int] = None,
    ) -> User:
        """
        Метод возвращает информацию о специалисте

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289421100/4.2.3.6.

        :param user_id: ID пользователя, являющегося специалистом по любой линии поддержки
        :param request_timeout: Request timeout
        :return: Return :code:`User`.
        """

        call = GetSpecialist()
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path=f"/v1/line/specialist/{is_valid_uuid(user_id)}/")

    async def get_specialists(
        self,
        request_timeout: Optional[int] = None,
    ) -> Users:
        """
        Метод возвращает информацию о специалистах

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2167373825/4.2.3.7.

        :param request_timeout: Request timeout
        :return: Return :code:`Users`.
        """

        call = GetSpecialists()
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path="/v1/line/specialists/")

    async def get_specialists_available(
        self,
        line_id: UUID,
        request_timeout: Optional[int] = None,
    ) -> List[UUID]:
        """
        Метод возвращает id специалистов в статусе "Доступен" по конкретной линии поддержки
        Вызов метода позволяет узнать можно ли назначить специалиста на текущий момент методом

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2167472129/4.2.3.8.

        :param line_id: ID линии поддержки
        :param request_timeout: Request timeout
        :return: Return :code:`Array[UUID]`.
        """

        call = GetSpecialistsAvailable()
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path=f"/v1/line/specialists/{is_valid_uuid(line_id)}/available/")

    async def get_competences(
        self,
        user_id: Optional[UUID] = None,
        line_id: Optional[UUID] = None,
        request_timeout: Optional[int] = None,
    ) -> Competences:
        """
        Метод возвращает список компетенций специалистов, с возможностью отфильтровать по специалисту и линии.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2490302465/4.2.3.9.

        :param user_id: ID пользователя (специалиста)
        :param line_id: ID линии поддержки
        :param request_timeout: Request timeout
        :return: Return :code:`Treatment`.
        """

        call = GetCompetences(
            user_id=is_valid_uuid(user_id),
            line_id=is_valid_uuid(line_id),
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path="/v1/line/competences/")

    async def get_ticket(
        self,
        id: UUID,
        request_timeout: Optional[int] = None,
    ) -> TicketShort:
        """
        Метод возвращает заявку Service Desk со вложенными объектами по id

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2492923976/4.2.3.10.+Service+Desk+ID

        :param id: ID заявки
        :param request_timeout: Request timeout
        :return: Return :code:`TicketShort`.
        """

        call = GetTicket()
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path=f"/v1/ticket/{is_valid_uuid(id)}/")

    async def get_ticket_by_number(
        self,
        number: int,
        request_timeout: Optional[int] = None,
    ) -> TicketShort:
        """
        Метод возвращает заявку Service Desk со вложенными объектами по номеру

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2492956727/4.2.3.11.+Service+Desk

        :param number: № заявки
        :param request_timeout: Request timeout
        :return: Return :code:`TicketShort`.
        """

        call = GetTicketByNumber()
        return await self(call, request_timeout=request_timeout,
                          type_request="GET", path=f"/v1/ticket/number/{number}/")

    """
    4.3.1. Команды внешних ботов

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1355841626/4.3.1.
    """

    async def appoint_start(
        self,
        line_id: UUID,
        user_id: UUID,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет назначить любого свободного специалиста.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1348632607/4.3.1.1.

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = AppointStart(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id),
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/line/appoint/start/")

    async def appoint_spec(
        self,
        line_id: UUID,
        user_id: UUID,
        spec_id: UUID,
        author_id: Optional[UUID] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет сделать попытку назначить конкретного специалиста.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1348632627/4.3.1.2.

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param spec_id: ID специалиста, на которого пытаемся назначить (должен быть в онлайне)
        :param author_id: ID автора (специалиста, от имени которого производится действие)
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = AppointSpec(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id),
            spec_id=is_valid_uuid(spec_id),
            author_id=is_valid_uuid(author_id),
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/line/appoint/spec/")

    async def drop_treatment(
        self,
        line_id: UUID,
        user_id: UUID,
        author_id: Optional[UUID] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет закрыть текущее обращение.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1347551359/4.3.1.3.

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param author_id: ID автора (специалиста, от имени которого производится действие)
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = DropTreatment(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id),
            author_id=is_valid_uuid(author_id),
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/line/drop/treatment/")

    async def send_message_line(
        self,
        line_id: UUID,
        user_id: UUID,
        text: str,
        author_id: Optional[UUID] = None,
        bot_as_spec: Optional[bool] = None,
        notification_only: Optional[bool] = None,
        keyboard: Optional[List[List[Button]]] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить сообщение в чат.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289355725/4.3.1.4.

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param text: Текст сообщения
        :param author_id: ID автора (специалиста, от имени которого отправляется сообщение)
        :param bot_as_spec: Флаг, при наличии которого не требуется получения ответов от пользователя
        (отправка клавиатуры запрещена), и может отсутствовать подписка на события с типом bot.
        Специалист, от лица которого отправляется сообщение (поле author_id), должен быть в онлайне и в статусе
        отличном от "Нет на месте" и "Не беспокоить", во избежании лишних переназначений или постановок пользователя в
        очередь ожидания специалиста.
        :param notification_only: Флаг, при наличии которого не требуется получения ответов от пользователя
        (отправка клавиатуры запрещена), и может отсутствовать подписка на события с типом bot. Используется для
        информирования пользователя о событиях. Специалист, от лица которого отправляется сообщение (поле author_id),
        не становится назначенным. При отсутствии открытого обращения, новое не будет открыто.
        :param keyboard: Клавиатура
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendMessageLine(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id),
            author_id=is_valid_uuid(author_id),
            text=text,
            bot_as_spec=bot_as_spec,
            notification_only=notification_only,
            keyboard=keyboard
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/line/send/message/")

    async def send_file_line(
        self,
        line_id: UUID,
        user_id: UUID,
        file_name: str,
        file: Union[InputFile, str],
        author_id: Optional[UUID] = None,
        comment: Optional[str] = None,
        bot_as_spec: Optional[bool] = None,
        notification_only: Optional[bool] = None,
        keyboard: Optional[List[List[Dict[Any, str]]]] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить файл в чат.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289388419/4.3.1.5.

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param file_name: Имя файла
        :param comment: Комментарий к файлу
        :param author_id: ID автора (специалиста, от имени которого отправляется сообщение)
        :param bot_as_spec: Флаг, при наличии которого не требуется получения ответов от пользователя
        (отправка клавиатуры запрещена), и может отсутствовать подписка на события с типом bot.
        Специалист, от лица которого отправляется сообщение (поле author_id), должен быть в онлайне и в статусе
        отличном от "Нет на месте" и "Не беспокоить", во избежании лишних переназначений или постановок пользователя в
        очередь ожидания специалиста.
        :param notification_only: Флаг, при наличии которого не требуется получения ответов от пользователя
        (отправка клавиатуры запрещена), и может отсутствовать подписка на события с типом bot. Используется для
        информирования пользователя о событиях. Специалист, от лица которого отправляется сообщение (поле author_id),
        не становится назначенным. При отсутствии открытого обращения, новое не будет открыто.
        :param keyboard: Клавиатура
        :param file: Файл
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendFileLine(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id),
            author_id=is_valid_uuid(author_id),
            file_name=file_name,
            comment=comment,
            bot_as_spec=bot_as_spec,
            notification_only=notification_only,
            keyboard=keyboard,
            file=file
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST-With-Attach", path="/v1/line/send/file/")

    async def send_image_line(
        self,
        line_id: UUID,
        user_id: UUID,
        file_name: str,
        file: Union[InputFile, str],
        author_id: Optional[UUID] = None,
        comment: Optional[str] = None,
        bot_as_spec: Optional[bool] = None,
        notification_only: Optional[bool] = None,
        keyboard: Optional[List[List[Dict[Any, str]]]] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить картинку в чат.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1492091047/4.3.1.6.

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param file_name: Имя файла
        :param comment: Комментарий к файлу
        :param author_id: ID автора (специалиста, от имени которого отправляется сообщение)
        :param bot_as_spec: Флаг, при наличии которого не требуется получения ответов от пользователя
        (отправка клавиатуры запрещена), и может отсутствовать подписка на события с типом bot.
        Специалист, от лица которого отправляется сообщение (поле author_id), должен быть в онлайне и в статусе
        отличном от "Нет на месте" и "Не беспокоить", во избежании лишних переназначений или постановок пользователя в
        очередь ожидания специалиста.
        :param notification_only: Флаг, при наличии которого не требуется получения ответов от пользователя
        (отправка клавиатуры запрещена), и может отсутствовать подписка на события с типом bot. Используется для
        информирования пользователя о событиях. Специалист, от лица которого отправляется сообщение (поле author_id),
        не становится назначенным. При отсутствии открытого обращения, новое не будет открыто.
        :param keyboard: Клавиатура
        :param file: Файл
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendImageLine(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id),
            author_id=is_valid_uuid(author_id),
            file_name=file_name,
            comment=comment,
            bot_as_spec=bot_as_spec,
            notification_only=notification_only,
            keyboard=keyboard,
            file=file
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST-With-Attach", path="/v1/line/send/image/")

    async def drop_keyboard(
        self,
        line_id: UUID,
        user_id: UUID,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет убрать клавиатуру.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1347551311/4.3.1.7.+-

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = DropKeyboard(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id)
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/line/drop/keyboard/")

    async def send_message_colleague(
        self,
        recepient_id: UUID,
        author_id: UUID,
        text: str,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить сообщение в чат сотруднику от лица другого сотрудника.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1479737345/4.3.1.8.

        :param recepient_id: ID получателя
        :param author_id: ID автора (от чьего имени отправляется сообщение)
        :param text: Текст сообщения
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendMessageColleague(
            recepient_id=is_valid_uuid(recepient_id),
            author_id=is_valid_uuid(author_id),
            text=text,
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/colleague/send/message/")

    async def send_file_collegue(
        self,
        recepient_id: UUID,
        author_id: UUID,
        file_name: str,
        file: Union[InputFile, str],
        comment: Optional[str] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить файл в чат сотруднику от имени другого сотрудника.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1479442525/4.3.1.9.

        :param recepient_id: ID получателя
        :param author_id: ID автора (от чьего имени отправляется файл)
        :param file_name: Имя файла
        :param comment: Комментарий к файлу
        :param file: Файл
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendFileCollegue(
            recepient_id=is_valid_uuid(recepient_id),
            author_id=is_valid_uuid(author_id),
            file_name=file_name,
            comment=comment,
            file=file
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST-With-Attach", path="/v1/colleague/send/file/")

    async def send_image_collegue(
        self,
        recepient_id: UUID,
        author_id: UUID,
        file_name: str,
        file: Union[InputFile, str],
        comment: Optional[str] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить картинку в чат сотруднику от имени другого сотрудника.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1492091064/4.3.1.10.

        :param recepient_id: ID получателя
        :param author_id: ID автора (от чьего имени отправляется файл)
        :param file_name: Имя файла
        :param comment: Комментарий к файлу
        :param file: Картинка
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendImageColleague(
            recepient_id=is_valid_uuid(recepient_id),
            author_id=is_valid_uuid(author_id),
            file_name=file_name,
            comment=comment,
            file=file
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST-With-Attach", path="/v1/colleague/send/image/")

    async def send_message_conference(
        self,
        conference_id: UUID,
        author_id: UUID,
        text: str,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить сообщение в группу от имени участника этой группы.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1479704633/4.3.1.11.

        :param conference_id: ID группы
        :param author_id: ID автора (от чьего имени отправляется сообщение)
        :param text: Текст сообщения
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendMessageConference(
            conference_id=is_valid_uuid(conference_id),
            author_id=is_valid_uuid(author_id),
            text=text,
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/conference/send/message/")

    async def send_file_conference(
        self,
        conference_id: UUID,
        author_id: UUID,
        file_name: str,
        file: Union[InputFile, str],
        comment: Optional[str] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить файл в группу от имени участника этой группы.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1478099061/4.3.1.12.

        :param conference_id: ID группы
        :param author_id: ID автора (от чьего имени отправляется файл)
        :param file_name: Имя файла
        :param comment: Комментарий к файлу
        :param file: Файл
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendFileConference(
            conference_id=is_valid_uuid(conference_id),
            author_id=is_valid_uuid(author_id),
            file_name=file_name,
            comment=comment,
            file=file
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST-With-Attach", path="/v1/conference/send/file/")

    async def send_image_conference(
        self,
        conference_id: UUID,
        author_id: UUID,
        file_name: str,
        file: Union[InputFile, str],
        comment: Optional[str] = None,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отправить картинку в группу от имени участника этой группы.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1492091105/4.3.1.13.

        :param conference_id: ID получателя
        :param author_id: ID автора (от чьего имени отправляется файл)
        :param file_name: Имя файла
        :param comment: Комментарий к файлу
        :param file: Картинка
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = SendImageConference(
            conference_id=is_valid_uuid(conference_id),
            author_id=is_valid_uuid(author_id),
            file_name=file_name,
            comment=comment,
            file=file
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST-With-Attach", path="/v1/conference/send/image/")

    async def question_and_answering(
        self,
        line_id: UUID,
        user_id: UUID,
        skip_greetings: bool,
        skip_goodbyes: bool,
        request_timeout: Optional[int] = None,
    ) -> Answering:
        """
        Метод позволяет получить варианты ответов на вопрос пользователя в базе знаний.

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1481932813/4.3.1.14.

        :param line_id: ID линии поддержки
        :param user_id: ID пользователя
        :param skip_greetings: Не использовать “Единая база приветствий и поддержки диалогов“
        :param skip_goodbyes: Не использовать “Единая база завершения диалогов“
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = QuestionAndAnswering(
            line_id=is_valid_uuid(line_id),
            user_id=is_valid_uuid(user_id),
            skip_greetings=skip_greetings,
            skip_goodbyes=skip_goodbyes,
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="POST", path="/v1/line/qna/")

    async def question_and_answering_selected(
        self,
        request_id: UUID,
        result_id: UUID,
        request_timeout: Optional[int] = None,
    ) -> bool:
        """
        Метод позволяет отметить конкретную подсказку выбранной

        Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2290024449

        :param request_id: ID запроса
        :param result_id: ID выбранной статьи
        :param request_timeout: Request timeout
        :return: Нет тела
        """

        call = QuestionAndAnsweringSelected(
            request_id=is_valid_uuid(request_id),
            result_id=is_valid_uuid(result_id),
        )
        return await self(call, request_timeout=request_timeout,
                          type_request="PUT", path="/v1/line/qna/selected/")



from .base import Request, Response, ConnectMethod
# 4.2.1 Команды к механизму трансляции
from .set_hook import SetHook
from .del_all_hook import DelAllHook
from .del_hook import DelHook
# 4.2.3 Команды для уточнения информации
from .get_treatments import GetTreatments
from .get_subscriber import GetSubscriber
from .get_subscribers import GetSubscribers
from .get_subscriptions import GetSubscriptions
from .get_lines import GetLines
from .get_specialist import GetSpecialist
from .get_specialists import GetSpecialists
from .get_specialists_available import GetSpecialistsAvailable
from .get_competences import GetCompetences
from .get_ticket import GetTicket
from .get_ticket_by_number import GetTicketByNumber
# 4.3.1. Команды внешних ботов
from .appoint_start import AppointStart
from .appoint_spec import AppointSpec
from .drop_treatment import DropTreatment
from .send_message_line import SendMessageLine
from .send_file_line import SendFileLine
from .send_image_line import SendImageLine
from .drop_keyboard import DropKeyboard
from .send_message_colleague import SendMessageColleague
from .send_file_collegue import SendFileCollegue
from .send_image_collegue import SendImageColleague
from .send_message_conference import SendMessageConference
from .send_file_conference import SendFileConference
from .send_image_conference import SendImageConference
from .question_and_answering import QuestionAndAnswering, QuestionAndAnsweringSelected

__all__ = (
    "Request", "Response", "ConnectMethod",
    # 4.2.1 Команды к механизму трансляции
    "SetHook",
    "DelAllHook",
    "DelHook",
    # 4.2.3 Команды для уточнения информации
    "GetTreatments",
    "GetSubscriber",
    "GetSubscribers",
    "GetSubscriptions",
    "GetLines",
    "GetSpecialist",
    "GetSpecialists",
    "GetSpecialistsAvailable",
    "GetCompetences",
    "GetTicket",
    "GetTicketByNumber",
    # 4.3.1. Команды внешних ботов
    "AppointStart",
    "AppointSpec",
    "DropTreatment",
    "SendMessageLine",
    "SendFileLine",
    "SendImageLine",
    "DropKeyboard",
    "SendMessageColleague",
    "SendFileCollegue",
    "SendImageColleague",
    "SendMessageConference",
    "SendFileConference",
    "SendImageConference",
    "QuestionAndAnswering", "QuestionAndAnsweringSelected"
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .base import ConnectMethod

from ..types import UUID, Answering


class QuestionAndAnswering(ConnectMethod[Answering]):
    """
    Типы запросов: POST
    Описание: Получение подсказок
    Название: QuestionAndAnswering
    URL: /v1/line/qna/

    Метод позволяет получить варианты ответов на вопрос пользователя в базе знаний.
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1481932813/4.3.1.14.
    """

    __returning__ = Answering

    line_id: UUID
    """ID линии поддержки"""
    user_id: UUID
    """ID пользователя"""
    skip_greetings: bool
    """Не использовать “Единая база приветствий и поддержки диалогов“"""
    skip_goodbyes: bool
    """Не использовать “Единая база завершения диалогов“"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            line_id: UUID,
            user_id: UUID,
            skip_greetings: bool,
            skip_goodbyes: bool,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                line_id=line_id,
                user_id=user_id,
                skip_greetings=skip_greetings,
                skip_goodbyes=skip_goodbyes,
                **__pydantic_kwargs,
            )


class QuestionAndAnsweringSelected(ConnectMethod[bool]):
    """
    Типы запросов: PUT
    Описание: Получение подсказок
    Название: QuestionAndAnsweringSelected
    URL: /v1/line/qna/selected/

    Метод позволяет отметить конкретную подсказку выбранной
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2290024449
    """

    __returning__ = bool

    request_id: UUID
    """ID запроса"""
    result_id: UUID
    """ID выбранной статьи"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            request_id: UUID,
            result_id: UUID,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                request_id=request_id,
                result_id=result_id,
                **__pydantic_kwargs,
            )



from typing import Any, Dict
from unittest.mock import sentinel

from pydantic import BaseModel, ConfigDict, model_validator

from ..client.context_controller import BotContextController


class ConnectObject(BotContextController, BaseModel):
    model_config = ConfigDict(
        use_enum_values=True,
        extra="allow",
        validate_assignment=True,
        frozen=True,
        populate_by_name=True,
        arbitrary_types_allowed=True,
        defer_build=True,
    )

    @model_validator(mode="before")
    @classmethod
    def remove_unset(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Remove UNSET before fields validation.

        We use UNSET as a sentinel value for `parse_mode` and replace it to real value later.
        It isn't a problem when it's just default value for a model field,
        but UNSET might be passed to a model initialization from `Bot.method_name`,
        so we must take care of it and remove it before fields validation.
        """
        if not isinstance(values, dict):
            return values
        return {k: v for k, v in values.items() if not isinstance(v, UNSET_TYPE)}


class MutableConnectObject(ConnectObject):
    model_config = ConfigDict(
        frozen=False,
    )


# special sentinel object which used in a situation when None might be a useful value
UNSET: Any = sentinel.UNSET
UNSET_TYPE: Any = type(UNSET)



from __future__ import annotations

from .base import ConnectMethod


class DelHook(ConnectMethod[bool]):
    """
    Типы запросов: DELETE
    Описание: Удаление адреса конкретного webhook
    Название: DelHook
    URL: /v1/hook/{type}/{id}/

    id: str = ID объекта
    type: str = Тип WebHook
                - bot - на события для чат бота
                - line - все события по линии(ям) поддержки

    Метод удаляет адрес конкретного webhook.
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289355682/4.2.1.3.
    """

    __returning__ = bool



import asyncio
from abc import ABC, abstractmethod
from typing import Any, Dict, Set

from aiohttp import web
from aiohttp.abc import Application

from .. import Bot, Dispatcher
from ..methods import ConnectMethod


def setup_application(app: Application, dispatcher: Dispatcher, /, **kwargs: Any) -> None:
    """
    This function helps to configure a startup-shutdown process

    :param app: aiohttp application
    :param dispatcher: aio-connect dispatcher
    :param kwargs: additional data
    :return:
    """
    workflow_data = {
        "app": app,
        "dispatcher": dispatcher,
        **dispatcher.workflow_data,
        **kwargs,
    }

    async def on_startup(*a: Any, **kw: Any) -> None:  # pragma: no cover
        await dispatcher.emit_startup(**workflow_data)

    async def on_shutdown(*a: Any, **kw: Any) -> None:  # pragma: no cover
        await dispatcher.emit_shutdown(**workflow_data)

    app.on_startup.append(on_startup)
    app.on_shutdown.append(on_shutdown)


class BaseRequestHandler(ABC):
    def __init__(
        self,
        dispatcher: Dispatcher,
        **data: Any,
    ) -> None:
        """
        Base handler that helps to handle incoming request from aiohttp
        and propagate it to the Dispatcher

        :param dispatcher: instance of :class:`aio_connect.dispatcher.dispatcher.Dispatcher`
        """
        self.dispatcher = dispatcher
        self.data = data
        self._background_feed_update_tasks: Set[asyncio.Task[Any]] = set()

    def register(self, app: Application, /, path: str, **kwargs: Any) -> None:
        """
        Register route and shutdown callback

        :param app: instance of aiohttp Application
        :param path: route path
        :param kwargs:
        """
        app.on_shutdown.append(self._handle_close)
        app.router.add_route("POST", path, self.handle, **kwargs)

    async def _handle_close(self, app: Application) -> None:
        await self.close()

    @abstractmethod
    async def close(self) -> None:
        pass

    @abstractmethod
    async def resolve_bot(self, request: web.Request) -> Bot:
        """
        This method should be implemented in subclasses of this class.

        Resolve Bot instance from request.

        :param request:
        :return: Bot instance
        """
        pass

    @abstractmethod
    def verify_bot(self, bot: Bot) -> bool:
        if bot.auth:
            return True

    async def _background_feed_update(self, bot: Bot, update: Dict[str, Any]) -> None:
        result = await self.dispatcher.feed_raw_update(bot=bot, update=update, **self.data)
        if isinstance(result, ConnectMethod):
            await self.dispatcher.silent_call_request(bot=bot, result=result)

    async def _handle_request_background(self, bot: Bot, request: web.Request) -> web.Response:
        update = await request.json(loads=bot.session.json_loads)
        new_update = {'event_type': update["event_type"], 'event_source': update["event_source"],
                      update["event_type"]: {obj: update[obj] for obj in update if
                                             obj not in ('event_type', 'event_source')}}
        print("------")  # Fixme: remove
        print(new_update)  # Fixme: remove
        feed_update_task = asyncio.create_task(
            self._background_feed_update(
                bot=bot, update=new_update
            )
        )
        self._background_feed_update_tasks.add(feed_update_task)
        feed_update_task.add_done_callback(self._background_feed_update_tasks.discard)
        return web.json_response({}, dumps=bot.session.json_dumps)

    async def handle(self, request: web.Request) -> web.Response:
        bot = await self.resolve_bot(request)
        if not self.verify_bot(bot):
            return web.Response(body="Unauthorized", status=401)
        return await self._handle_request_background(bot=bot, request=request)

    __call__ = handle


class SimpleRequestHandler(BaseRequestHandler):
    def __init__(
        self,
        dispatcher: Dispatcher,
        bot: Bot,
        **data: Any,
    ) -> None:
        """
        Handler for single Bot instance

        :param dispatcher: instance of :class:`aio_connect.dispatcher.dispatcher.Dispatcher`
        :param bot: instance of :class:`aio_connect.client.bot.Bot`
        """
        super().__init__(dispatcher=dispatcher, **data)
        self.bot = bot

    def verify_bot(self, bot: Bot) -> bool:
        if bot.auth:
            return True

    async def close(self) -> None:
        """
        Close bot session
        """
        await self.bot.session.close()

    async def resolve_bot(self, request: web.Request) -> Bot:
        return self.bot



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .base import ConnectObject


class TicketAdditionalFieldValue(ConnectObject):
    """
    Название объекта: TicketAdditionalFieldValue
    Описание объекта: Дополнительное поле заявки Service Desk

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2492923923/TicketAdditionalFieldValue
    """

    id: str
    """Идентификатор поля:
        FIELD1 - Первое
        FIELD2 - Второе
        FIELD3 - Третье
        FIELD4 - Четвертое
    """
    value: str
    """Значение поля"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            id: str,
            value: str,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                id=id,
                value=value,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from datetime import datetime
from ..types import UUID


class Treatment(ConnectObject):
    """
    Название объекта: Treatment
    Описание объекта: Обращение

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289322810/Treatment
    """

    treatment_id: UUID
    """ID обращения"""
    line_id: UUID
    """ID линии поддержки"""
    user_id: UUID
    """ID пользователя"""
    initialized_at: datetime
    """Дата начала обращения"""
    treatment_duration: Optional[int] = None
    """Общая длительность обращения, сек. (при закрытии)"""
    current_specialist: Optional[UUID] = None
    """ID назначенного специалиста, null когда нет назначенного специалиста"""
    quality: Optional[int] = None
    """Оценка качества работы специалиста"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            treatment_id: UUID,
            line_id: UUID,
            user_id: UUID,
            initialized_at: datetime,
            treatment_duration: Optional[int] = None,
            current_specialist: Optional[UUID] = None,
            quality: Optional[int] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                treatment_id=treatment_id,
                line_id=line_id,
                user_id=user_id,
                initialized_at=initialized_at,
                treatment_duration=treatment_duration,
                current_specialist=current_specialist,
                quality=quality,
                **__pydantic_kwargs,
            )


class Treatments(ConnectObject):
    """
    Название объекта: Treatments
    Описание объекта: Обращения

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2492923941/Treatments
    """

    treatment_id: UUID
    """ID обращения"""
    line_id: UUID
    """ID линии поддержки"""
    user_id: UUID
    """ID пользователя"""
    initialized_at: datetime
    """Дата начала обращения"""
    treatment_duration: Optional[int] = None
    """Общая длительность обращения, сек. (при закрытии)"""
    current_specialist: Optional[UUID] = None
    """ID назначенного специалиста, null когда нет назначенного специалиста"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            treatment_id: UUID,
            line_id: UUID,
            user_id: UUID,
            initialized_at: datetime,
            treatment_duration: Optional[int] = None,
            current_specialist: Optional[UUID] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                treatment_id=treatment_id,
                line_id=line_id,
                user_id=user_id,
                initialized_at=initialized_at,
                treatment_duration=treatment_duration,
                current_specialist=current_specialist,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, List, Dict, Union

from .base import ConnectMethod
from ..types import UUID, InputFile


class SendFileLine(ConnectMethod[bool]):
    """
    Типы запросов: POST
    Описание: Отправка файла
    Название: SendFileLine
    URL: /v1/line/send/file/

    Метод позволяет отправить файл в чат.
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289388419/4.3.1.5.
    """

    __returning__ = bool

    line_id: UUID
    """ID линии поддержки"""
    user_id: UUID
    """ID пользователя"""
    author_id: Optional[UUID] = None
    """ID автора (специалиста, от имени которого отправляется файл)"""
    file_name: str
    """Имя файла"""
    comment: Optional[str] = None
    "Комментарий к файлу"
    bot_as_spec: Optional[bool] = None
    """Флаг, при наличии которого не требуется получения ответов от пользователя (отправка клавиатуры запрещена),
    и может отсутствовать подписка на события с типом bot.
    Специалист, от лица которого отправляется файл (поле author_id), должен быть в онлайне и в статусе отличном от
    "Нет на месте" и "Не беспокоить", во избежании лишних переназначений или постановок пользователя в очередь
    ожидания специалиста.
    """
    notification_only: Optional[bool] = None
    """Флаг, при наличии которого не требуется получения ответов от пользователя (отправка клавиатуры запрещена),
    и может отсутствовать подписка на события с типом bot.
    Используется для информирования пользователя о событиях.
    Специалист, от лица которого отправляется сообщение (поле author_id), не становится назначенным. При отсутствии
    открытого обращения, новое не будет открыто.
    """
    keyboard: Optional[List[List[Dict[Any, str]]]] = None
    """Клавиатура"""

    file: Union[InputFile, str]
    """Файл в виде:
        URLInputFile(url="{url_ссылка_на_файл}")
        FSInputFile(path="{./путь_к_файлу}")
        BufferedInputFile(file=b"{содержимое}, filename="{название_файла}")
    """

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            line_id: UUID,
            user_id: UUID,
            author_id: Optional[UUID] = None,
            file_name: str,
            comment: Optional[str] = None,
            bot_as_spec: Optional[bool] = None,
            notification_only: Optional[bool] = None,
            keyboard: Optional[List[List[Dict[Any, str]]]] = None,
            file: Union[InputFile, str],
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                line_id=line_id,
                user_id=user_id,
                author_id=author_id,
                file_name=file_name,
                comment=comment,
                bot_as_spec=bot_as_spec,
                notification_only=notification_only,
                keyboard=keyboard,
                file=file,
                **__pydantic_kwargs,
            )



from typing import Any, Awaitable, Callable, Dict, Optional, Tuple

from .base import BaseMiddleware
from ...types import ConnectObject, Update, Line, User

EVENT_LINE_ID = "event_line_id"
EVENT_USER_ID = "event_user_id"
EVENT_AUTHOR_ID = "event_author_id"
EVENT_ACTION = "event_action"


class UserContextMiddleware(BaseMiddleware):
    async def __call__(
        self,
        handler: Callable[[ConnectObject, Dict[str, Any]], Awaitable[Any]],
        event: ConnectObject,
        data: Dict[str, Any],
    ) -> Any:
        if not isinstance(event, Update):
            raise RuntimeError("UserContextMiddleware got an unexpected event type!")
        line_id, user_id, author_id, action = self.resolve_event_context(event=event)
        if line_id is not None:
            data[EVENT_LINE_ID] = line_id
        if user_id is not None:
            data[EVENT_USER_ID] = user_id
        if author_id is not None:
            data[EVENT_AUTHOR_ID] = author_id
        if action is not None:
            data[EVENT_ACTION] = action
        return await handler(event, data)

    @classmethod
    def resolve_event_context(
        cls, event: Update
    ) -> Tuple[Optional[Line], Optional[User], Optional[User], Optional[str]]:
        """
        Resolve chat and user instance from Update object
        """
        if event.event_type == "competence":
            return None, None, None, event.competence.action
        if event.event_type == "line":
            return event.line.line_id, event.line.user_id, event.line.author_id, None
        if event.event_type == "subscriber":
            return None, None, None, event.subscriber.action
        if event.event_type == "subscription":
            return None, None, None, event.subscription.action
        if event.event_type == "support_line":
            return None, None, None, event.support_line.action

        return None, None, None, None



from __future__ import annotations

from .base import ConnectMethod

from ..types import User


class GetSpecialist(ConnectMethod[User]):
    """
    Типы запросов: GET
    Описание: Получение информации о специалисте
    Название: GetSpecialist
    URL: /v1/line/specialist/{user_id}/

    user_id: UUID = ID пользователя, являющегося специалистом по любой линии поддержки

    Метод возвращает информацию о специалисте
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289421100/4.2.3.6.
    """

    __returning__ = User



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectMethod

from ..types import UUID, HookType


class SetHook(ConnectMethod[bool]):
    """
    Типы запросов: POST
    Описание: Выставление адреса webhook
    Название: SetHook
    URL: /v1/hook/

    Метод устанавливает webhook для получения событий.
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289388306/4.2.1.1.
    """

    __returning__ = bool

    url: str
    """URL WebHook. На этот адрес будут прилетать все события POST-запросами."""
    type: HookType
    """
    Тип подписки:
        - bot - на события для чат бота для указанной линии поддержки (обязателен id)
        - line - все события по линии(ям) поддержки
    """
    id: Optional[UUID] = None
    """
    ID объекта, на который осуществляется подписка:
        ID линии поддержки
        если type == line и не указан id, то подписка производится на все линии, к которым имеется доступ
    """

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            url: str,
            type: HookType,
            id: Optional[UUID] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                url=url,
                type=type,
                id=id,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from ..types import UUID


class Data(ConnectObject):
    """
    Название объекта: Data
    Описание объекта: Дополнительные данные

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1487241255/Data
    """

    redirect: Optional[str] = None
    """Дополнительные данные перевода на бота"""
    direction: Optional[str] = None
    """Направления перевода обращения
        from - откуда
        to - куда
        to_fran - перевод в компанию франа
        to_vendor - перевод в компанию вендора
    """
    line_id: Optional[UUID] = None
    """ID линии поддержки при direction:
        from - ID исходной линии поддержки
        to - ID линии поддержки назначения
    """
    treatment_id: Optional[UUID] = None
    """ID обращения при direction:
        from - ID исходного обращения
        to - ID нового обращения
    """
    company_id: Optional[UUID] = None
    """ID компании при direction:
         to_fran - франа
         to_vendor - вендора
    """

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            redirect: Optional[str] = None,
            direction: Optional[str] = None,
            line_id: Optional[UUID] = None,
            treatment_id: Optional[UUID] = None,
            company_id: Optional[UUID] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                redirect=redirect,
                direction=direction,
                line_id=line_id,
                treatment_id=treatment_id,
                company_id=company_id,
                **__pydantic_kwargs,
            )



from typing import Any, Optional

from .methods import ConnectMethod
from .methods.base import ConnectType


class AioconnectError(Exception):
    """
    Base exception for all Aio-connect errors.
    """


class DetailedAioconnectError(AioconnectError):
    """
    Base exception for all Aio-connect errors with detailed message.
    """

    url: Optional[str] = None

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        message = self.message
        if self.url:
            message += f"\n(background on this error at: {self.url})"
        return message

    def __repr__(self) -> str:
        return f"{type(self).__name__}('{self}')"


class UnsupportedKeywordArgument(DetailedAioconnectError):
    """
    Exception raised when a keyword argument is passed as filter.
    """


class ConnectAPIError(DetailedAioconnectError):
    """
    Base exception for all Connect API errors.
    """

    label: str = "Server says"

    def __init__(
        self,
        method: ConnectMethod[ConnectType],
        message: Optional[str]
    ) -> None:
        super().__init__(message=message)
        self.method = method

    def __str__(self) -> str:
        original_message = super().__str__()
        return f"{self.label} - {original_message}"


class ConnectNetworkError(ConnectAPIError):
    """
    Base exception for all Connect network errors.
    """

    label = "HTTP Client says"


class ConnectBadRequest(ConnectAPIError):
    """
    400 - Запрос содержит ошибку
    """


class ConnectUnauthorizedError(ConnectAPIError):
    """
    401 - Пользователь не авторизован
    """


class ConnectForbiddenError(ConnectAPIError):
    """
    403 - Недостаточно прав
    """


class ConnectNotFound(ConnectAPIError):
    """
    404 -
    """


class UnprocessalbleEntity(ConnectAPIError):
    """
    422 - Ошибка сервиса или запрос не прошел по требованиям
    """


class ConnectConflictError(ConnectAPIError):
    """
    409 -
    """


class ConnectServerError(ConnectAPIError):
    """
    Exception raised when Connect server returns 5xx error.
    """


class RestartingConnect(ConnectServerError):
    """
    Exception raised when Connect server is restarting.

    It seems like this error is not used by Connect anymore,
    but it's still here for backward compatibility.

    Currently, you should expect that Connect can raise RetryAfter (with timeout 5 seconds)
     error instead of this one.
    """


class ConnectEntityTooLarge(ConnectNetworkError):
    """
    Exception raised when you are trying to send a file that is too large.
    """


class ClientDecodeError(AioconnectError):
    """
    Exception raised when client can't decode response. (Malformed response, etc.)
    """

    def __init__(self, message: str, original: Exception, data: Any) -> None:
        self.message = message
        self.original = original
        self.data = data

    def __str__(self) -> str:
        original_type = type(self.original)
        return (
            f"{self.message}\n"
            f"Caused from error: "
            f"{original_type.__module__}.{original_type.__name__}: {self.original}\n"
            f"Content: {self.data}"
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Protocol

from ....methods import Response, ConnectMethod
from ....methods.base import ConnectType

if TYPE_CHECKING:
    from ...bot import Bot


class NextRequestMiddlewareType(Protocol[ConnectType]):  # pragma: no cover
    async def __call__(
        self,
        bot: "Bot",
        method: ConnectMethod[ConnectType],
    ) -> Response[ConnectType]:
        pass


class RequestMiddlewareType(Protocol):  # pragma: no cover
    async def __call__(
        self,
        make_request: NextRequestMiddlewareType[ConnectType],
        bot: "Bot",
        method: ConnectMethod[ConnectType],
    ) -> Response[ConnectType]:
        pass


class BaseRequestMiddleware(ABC):
    """
    Generic middleware class
    """

    @abstractmethod
    async def __call__(
        self,
        make_request: NextRequestMiddlewareType[ConnectType],
        bot: "Bot",
        method: ConnectMethod[ConnectType],
    ) -> Response[ConnectType]:
        """
        Execute middleware

        :param make_request: Wrapped make_request in middlewares chain
        :param bot: bot for request making
        :param method: Request method (Subclass of :class:`aio_connect.methods.base.ConnectMethod`)

        :return: :class:`aio_connect.methods.Response`
        """
        pass



from __future__ import annotations

from .base import ConnectMethod

from ..types import Users


class GetSubscribers(ConnectMethod[Users]):
    """
    Типы запросов: GET
    Описание: Получение информации о пользователях
    Название: GetSubscribers
    URL: /v1/line/subscribers/

    Метод возвращает информацию о пользователях
    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2068938753/4.2.3.3.
    """

    __returning__ = Users



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from ..types import UUID


class User(ConnectObject):
    """
    Название объекта: User
    Описание объекта: Пользователь

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1289355329/User
    """

    user_id: UUID
    """ID пользователя"""
    name: str
    """Имя пользователя"""
    surname: str
    """Фамилия пользователя"""
    patronymic: str
    """Отчество пользователя"""
    avatar_url: Optional[str] = None
    """Ссылка на аватар"""
    avatar_small_url: Optional[str] = None
    """Ссылка на превью аватара"""
    email: str
    """Email пользователя"""
    post: str
    """Должность пользователя"""
    phone: str
    """Телефон пользователя"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            user_id: UUID,
            name: str,
            surname: str,
            patronymic: str,
            avatar_url: Optional[str] = None,
            avatar_small_url: Optional[str] = None,
            email: str,
            post: str,
            phone: str,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                user_id=user_id,
                name=name,
                surname=surname,
                patronymic=patronymic,
                avatar_url=avatar_url,
                avatar_small_url=avatar_small_url,
                email=email,
                post=post,
                phone=phone,
                **__pydantic_kwargs,
            )


class Users(User):
    """
    Название объекта: Users
    Описание объекта: Пользователи

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2068938769/Users

    Данные: Такие же как у User
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .base import ConnectObject
from ..enums import ContentType

from ..types import UserServiceLine


class TypeSubscription(ConnectObject):
    """
    Дополнительные данные в сообщении:

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1357152269/4.2.2.2.4.+subscription
    """

    action: str
    """Тип изменения объектов:
        add
        update
        delete
    """
    subscription: UserServiceLine
    """Описание подписки"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            action: str,
            subscription: UserServiceLine,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                action=action,
                subscription=subscription,
                **__pydantic_kwargs,
            )

    @property
    def content_type(self) -> str:

        if self.action:
            return ContentType.ACTION
        if self.subscription:
            return ContentType.SUBSCRIPTION

        return ContentType.UNKNOWN



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .base import ConnectObject

from ..types import UUID


class TicketStatus(ConnectObject):
    """
    Название объекта: TicketStatus
    Описание объекта: Статусы заявок Service Desk

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/2492923906/TicketStatus
    """

    id: UUID
    """ID Статуса заявки"""
    name: str
    """Название"""
    type: str
    """Тип:
        NEW - Новая
        WORK - Выполняется
        VALIDATION - Валидация
        CONFIRMED - Подтверждена
        REJECTED - Отклонена
        SUSPENDED - Приостановлена
        FINISHED - Завершена
        CANCELLED - Отменена
    """

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            id: UUID,
            name: str,
            type: str,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                id=id,
                name=name,
                type=type,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List

from .base import ConnectObject

from ..types import UUID


class Answering(ConnectObject):
    """
    Название объекта: Answering
    Описание объекта: Ответ на вопрос
    """

    question: str
    """Текст вопроса, на который пытался дать варианты"""
    request_id: UUID
    """ID запроса"""
    answers: List
    """Массив с вариантами ответов. Отсортированы по уровню уверенности"""
    id: UUID
    """ID статьи базы знаний"""
    text: str
    """Текст ответа"""
    accuracy: float
    """Уровень уверенности(max. 1.0 - 100 %)"""
    answer_source: str
    """База знаний, из которой получен ответ:
        GREETINGS - приветствия
        GOODBYES - прощания
        OURS - любая из ваших
    """

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            question: str,
            request_id: UUID,
            answers: List,
            id: UUID,
            text: str,
            accuracy: float,
            answer_source: str,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                question=question,
                request_id=request_id,
                answers=answers,
                id=id,
                text=text,
                accuracy=accuracy,
                answer_source=answer_source,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import MutableConnectObject


class Button(MutableConnectObject):
    """
    Название объекта: Button
    Описание объекта: Структура описывающаю кнопку в бот-меню.

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1357119519/4.3.2.1.+Button
    """

    id: Optional[str] = None
    """Идентификатор команды"""
    text: str
    """Название команды"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            id: Optional[str] = None,
            text: str,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                id=id,
                text=text,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject

from ..types import UUID


class Rda(ConnectObject):
    """
    Название объекта: Rda
    Описание объекта: Сеанс удаленного доступа

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1304002561/Rda
    """

    src: UUID
    """ID инициатора"""
    dst: UUID
    """ID адресата"""
    duration: Optional[int] = None
    """Длительность сеанса, сек."""
    download_count: Optional[int] = None
    """Принятые файлы, шт."""
    upload_count: Optional[int] = None
    """Отправленные файлы, шт."""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
            __pydantic__self__,
            *,
            src: UUID,
            dst: UUID,
            duration: Optional[int] = None,
            download_count: Optional[int] = None,
            upload_count: Optional[int] = None,
            **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                src=src,
                dst=dst,
                duration=duration,
                download_count=download_count,
                upload_count=upload_count,
                **__pydantic_kwargs,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from .base import ConnectObject
from ..enums import ContentType

from datetime import datetime
from ..types import UUID, File, Call, Rda, ServiceRequest, Treatment, Data, PartnerNotification


class TypeLine(ConnectObject):
    """
    Дополнительные данные в сообщении:

    Source: https://1c-connect.atlassian.net/wiki/spaces/PUBLIC/pages/1357053953/4.2.2.2.2+line
    """

    message_id: UUID
    """ID сообщения"""
    message_type: int
    """Тип сообщения"""
    message_time: datetime
    """Время отправки сообщения"""
    treatment_id: Optional[UUID] = None
    """ID обращения
        Если это поле содержит неизвестный до текущего момента ID, то считается что начато новое обращение.
        Это справедливо для сообщений с любым типом.
    """
    author_id: Optional[UUID] = None
    """ID пользователя (может отсутствовать в случае инициации события автоматического закрытия обращения)"""
    line_id: UUID
    """ID линии поддержки"""
    user_id: UUID
    """ID пользователя"""
    text: Optional[str] = None
    """Текст сообщения"""
    file: Optional[File] = None
    """Данные файла"""
    call: Optional[Call] = None
    """Данные о звонке"""
    rda: Optional[Rda] = None
    """Данные о сеансе удаленного доступа"""
    service_request: Optional[ServiceRequest] = None
    """Данные заявки Service Desk"""
    treatment: Optional[Treatment] = None
    """Информация об обращении"""
    data: Optional[Data] = None
    """Дополнительные данные"""
    partner_notification: Optional[PartnerNotification] = None
    """Рассылка"""

    if TYPE_CHECKING:
        # DO NOT EDIT MANUALLY!!!
        # This section was auto-generated via `butcher`

        def __init__(
                __pydantic__self__,
                *,
                message_id: UUID,
                message_type: int,
                message_time: datetime,
                treatment_id: Optional[UUID] = None,
                author_id: Optional[UUID] = None,
                line_id: UUID,
                user_id: UUID,
                text: Optional[str] = None,
                file: Optional[File] = None,
                call: Optional[Call] = None,
                rda: Optional[Rda] = None,
                service_request: Optional[ServiceRequest] = None,
                treatment: Optional[Treatment] = None,
                data: Optional[Data] = None,
                partner_notification: Optional[PartnerNotification] = None,
                **__pydantic_kwargs: Any,
        ) -> None:
            # DO NOT EDIT MANUALLY!!!
            # This method was auto-generated via `butcher`
            # Is needed only for type checking and IDE support without any additional plugins

            super().__init__(
                message_id=message_id,
                message_type=message_type,
                message_time=message_time,
                treatment_id=treatment_id,
                author_id=author_id,
                line_id=line_id,
                user_id=user_id,
                text=text,
                file=file,
                call=call,
                rda=rda,
                service_request=service_request,
                treatment=treatment,
                data=data,
                partner_notification=partner_notification,
                **__pydantic_kwargs,
            )

    @property
    def content_type(self) -> str:

        TypeMessage = {
            1: ContentType.TEXT,  # Текстовое сообщение
            2: ContentType.NOTIFY,  # Рассылка

            17: ContentType.QUALITY_WORK,  # Оценка работы специалиста

            20: ContentType.CALL_START_WITH_NSS,  # Начало звонка с открытием обращения
            21: ContentType.CALL_START_WITHOUT_NSS,  # Начало звонка без открытия обращения
            22: ContentType.CALL_END_GOOD,  # Удачное завершение звонка
            23: ContentType.CALL_BADCALL_CANCELED,  # Неудачное завершение звонка. Вызов отменен инициатором
            24: ContentType.CALL_BADCALL_EXCESS,  # Неудачное завершение звонка. Абонент не ответил
            25: ContentType.CALL_BADCALL_NONSS,  # Неудачное завершение звонка. Нет свободных специалистов
            26: ContentType.CALL_BADCALL_REJECTED,  # Неудачное завершение звонка. Вызов отклонен адресатом
            27: ContentType.CALL_REROUTING,  # Перевод звонка на специалиста
            28: ContentType.CALL_REROUTING_VENDOR,  # Перевод звонка в компанию вендора
            30: ContentType.CALL_REROUTING_WITHOUTEND,  # Перевод звонка без завершения(когда абонент не взял
            # трубку, а сразу перевел звонок)
            31: ContentType.CALL_REROUTING_VENDORFRAN_WE,  # Перевод звонка без завершения на вендора(когда абонент
            # не взял трубку, а сразу перевел звонок)

            32: ContentType.CALL_TECHNICAL_PROBLEM,  # Нет соединения с голосовым сервером
            36: ContentType.CALL_TECHNICAL_PROBLEM_NO_AUDIO,  # Неудачное завершение. У абонента нет аудио
            # устройства
            38: ContentType.LINE_CALLUNAVAIL,  # Недоступная линия по звонку (нерабочее время)

            50: ContentType.RDA_START_WITH_NSS,  # Начало сеанса удаленного доступа с сервисным специалистом
            51: ContentType.RDA_START_WITHOUT_NSS,  # Начало сеанса удаленного доступа без сервисного специалиста
            # (Специалист принудительно назначается пользователю сеансом удаленного доступа)
            52: ContentType.RDA_END_WITHT_RANSFER_FILES,  # Окончание сеанса удаленного доступа с передачей файлов
            53: ContentType.RDA_END_WITHTOUT_RANSFER_FILES,  # Окончание сеанса удаленного доступа без передачей
            # файлов
            54: ContentType.RDA_BAD_CANCELED,  # Неудачное завершение удаленного доступа. Сеанс отменен инициатором
            55: ContentType.RDA_BAD_REJECTED,  # Неудачное завершение удаленного доступа. Сеанс отклонен принимающим
            56: ContentType.RDA_BAD_EXCESS,  # Неудачное завершение удаленного доступа. Сеанс не состоялся по
            # таймауту
            57: ContentType.RDA_BAD,  # Неудачное завершение удаленного доступа
            59: ContentType.RDA_BAD_OLD_SERVICE,  # Неудачное завершение - устаревшая версия службы
            60: ContentType.RDA_BAD_NO_SERVICE,  # Неудачное завершение - служба не установлена
            61: ContentType.RDA_BAD_OLD_COMPONENT,  # Неудачное завершение - устаревшие версии компонентов
            62: ContentType.RDA_BAD_NO_FILES,  # Неудачное завершение - отсутствуют файлы компонентов УД

            70: ContentType.TRANSFERFILES,  # Передача файла через чат

            80: ContentType.LINE_USERINIT,  # Назначение специалиста системой. Инициатором был пользователь
            81: ContentType.LINE_SPECINIT,  # Специалист назначился. Инициатором был специалист
            82: ContentType.LINE_SPECDEL,  # Завершение работы специалиста (Закрытие обращения)
            83: ContentType.LINE_NONSS,  # Обращение поступило в очередь. Нет свободных специалистов

            84: ContentType.LINE_REROUTINGSPEC,  # Перевод обращения на специалиста

            85: ContentType.LINE_REROUTING_VENDOR,  # Перевод обращения в компанию вендора

            86: ContentType.LINE_SPECFOUND,  # Для обращения в очереди появился свободный специалист
            87: ContentType.LINE_CHATUNAVAIL,  # Недоступность линии (нерабочее время)
            88: ContentType.LINE_REROUTE_UNAVAIL,  # Недоступность линии по переводу. При переводе обращения в
            # компанию вендора попали в нерабочее время

            89: ContentType.LINE_REROUTING_OTHERSERVICE,  # Перевод в другую линию поддержки

            90: ContentType.LINE_CLOSED_NO_ACTIVITY,  # Обращение закрыто автоматически по отсутствию активности в
            # чате
            91: ContentType.LINE_CLOSED_REMOVE_SERVICE,  # Обращение закрыто автоматически, т.к.удалена линия
            # поддержки
            92: ContentType.LINE_CLOSED_REMOVE_SUBSCRIPTION,  # Обращение закрыто автоматически, т.к.удалена
            # подписка пользователя
            93: ContentType.LINE_CLOSED_REMOVE_USER,  # Обращение закрыто автоматически, т.к.удален пользователь

            121: ContentType.SERVICE_REQUEST_ADD,  # Создание заявки Service Desk
            122: ContentType,  # Изменение заявки Service Desk
            123: ContentType,  # Завершение заявки Service Desk
            124: ContentType,  # Отмена заявки Service Desk

            200: ContentType.LINE_REROUTING_TO_BOT,  # Перевод обращения специалистом на бота
        }

        if self.message_type:
            return TypeMessage[self.message_type]

        return ContentType.UNKNOWN



from typing import Any, Awaitable, Callable, Dict, Optional

from ..dispatcher.middlewares.base import BaseMiddleware
from .context import FSMContext
from .storage.base import (
    DEFAULT_DESTINY,
    BaseEventIsolation,
    BaseStorage,
    StorageKey,
)
from ..types import ConnectObject, UUID


class FSMContextMiddleware(BaseMiddleware):
    def __init__(
        self,
        storage: BaseStorage,
        events_isolation: BaseEventIsolation,
    ) -> None:
        self.storage = storage
        self.events_isolation = events_isolation

    async def __call__(
        self,
        handler: Callable[[ConnectObject, Dict[str, Any]], Awaitable[Any]],
        event: ConnectObject,
        data: Dict[str, Any],
    ) -> Any:
        context = self.resolve_event_context(data)
        data["fsm_storage"] = self.storage
        if context:
            # Bugfix:
            # State should be loaded after lock is acquired
            async with self.events_isolation.lock(key=context.key):
                data.update({"state": context, "raw_state": await context.get_state()})
                return await handler(event, data)
        return await handler(event, data)

    def resolve_event_context(
        self,
        data: Dict[str, Any],
        destiny: str = DEFAULT_DESTINY,
    ) -> Optional[FSMContext]:
        line_id = data.get("event_line_id")
        user_id = data.get("event_user_id")
        author_id = data.get("event_author_id")
        action = data.get("event_action")
        return self.resolve_context(
            line_id=line_id,
            user_id=user_id,
            author_id=author_id,
            action=action,
            destiny=destiny,
        )

    def resolve_context(
        self,
        line_id: Optional[UUID] = None,
        user_id: Optional[UUID] = None,
        author_id: Optional[UUID] = None,
        action: Optional[str] = None,
        destiny: str = DEFAULT_DESTINY,
    ) -> Optional[FSMContext]:
        return self.get_context(
                line_id=line_id,
                user_id=user_id,
                author_id=author_id,
                action=action,
                destiny=destiny,
        )

    def get_context(
        self,
        line_id: Optional[UUID] = None,
        user_id: Optional[UUID] = None,
        author_id: Optional[UUID] = None,
        action: Optional[str] = None,
        destiny: str = DEFAULT_DESTINY,
    ) -> FSMContext:
        return FSMContext(
            storage=self.storage,
            key=StorageKey(
                line_id=line_id,
                user_id=user_id,
                author_id=author_id,
                action=action,
                destiny=destiny,
            ),
        )

    async def close(self) -> None:
        await self.storage.close()
        await self.events_isolation.close()



from __future__ import annotations

import abc
import datetime
import json
import secrets
from enum import Enum
from http import HTTPStatus
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Final,
    Optional,
    Type,
    cast,
)

from pydantic import ValidationError

from ...exceptions import (
    ClientDecodeError,
    RestartingConnect,
    ConnectAPIError,
    ConnectBadRequest,
    ConnectConflictError,
    ConnectEntityTooLarge,
    ConnectForbiddenError,
    ConnectNotFound,
    ConnectServerError,
    ConnectUnauthorizedError, UnprocessalbleEntity,
)

from .middlewares.manager import RequestMiddlewareManager
from ...methods import Response, ConnectMethod
from ...methods.base import ConnectType
from ...types import InputFile

if TYPE_CHECKING:
    from ..bot import Bot

_JsonLoads = Callable[..., Any]
_JsonDumps = Callable[..., str]

DEFAULT_TIMEOUT: Final[float] = 60.0


class BaseSession(abc.ABC):
    """
    This is base class for all HTTP sessions in aio-connect.

    If you want to create your own session, you must inherit from this class.
    """

    def __init__(
        self,
        json_loads: _JsonLoads = json.loads,
        json_dumps: _JsonDumps = json.dumps,
        timeout: float = DEFAULT_TIMEOUT,
    ) -> None:
        """

        :param json_loads: JSON loader
        :param json_dumps: JSON dumper
        :param timeout: Session scope request timeout
        """
        self.json_loads = json_loads
        self.json_dumps = json_dumps
        self.timeout = timeout

        self.middleware = RequestMiddlewareManager()

    def check_response(
        self, bot: Bot, method: ConnectMethod[ConnectType], status_code: int, content: str
    ) -> Response[Any]:
        """
        Check response status
        """
        try:
            json_data = {'ok': False, 'result': False, 'error_code': None}
            if status_code == HTTPStatus.OK:
                json_data['ok'] = True
                json_data['result'] = True
            else:
                json_data['error_code'] = status_code
            if content:
                json_data['result'] = self.json_loads(content)
        except Exception as e:
            # Handled error type can't be classified as specific error
            # in due to decoder can be customized and raise any exception
            raise ClientDecodeError("Failed to decode object", e, content)

        try:
            response_type = Response[Any]  # type: ignore
            response = response_type.model_validate(json_data, context={"bot": bot})
        except ValidationError as e:
            raise ClientDecodeError("Failed to deserialize object", e, json_data)

        if HTTPStatus.OK <= status_code <= HTTPStatus.IM_USED and response.ok:
            return response

        result = cast(str, response.result)

        if status_code == HTTPStatus.BAD_REQUEST:
            raise ConnectBadRequest(method=method, message=result)
        if status_code == HTTPStatus.NOT_FOUND:
            raise ConnectNotFound(method=method, message=result)
        if status_code == HTTPStatus.CONFLICT:
            raise ConnectConflictError(method=method, message=result)
        if status_code == HTTPStatus.UNAUTHORIZED:
            raise ConnectUnauthorizedError(method=method, message=result)
        if status_code == HTTPStatus.FORBIDDEN:
            raise ConnectForbiddenError(method=method, message=result)
        if status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            raise UnprocessalbleEntity(method=method, message=result)
        if status_code == HTTPStatus.REQUEST_ENTITY_TOO_LARGE:
            raise ConnectEntityTooLarge(method=method, message=result)
        if status_code >= HTTPStatus.INTERNAL_SERVER_ERROR:
            if "restart" in result:
                raise RestartingConnect(method=method, message=result)
            raise ConnectServerError(method=method, message=result)

        raise ConnectAPIError(
            method=method,
            message=result,
        )

    @abc.abstractmethod
    async def close(self) -> None:  # pragma: no cover
        """
        Close client session
        """
        pass

    @abc.abstractmethod
    async def make_request(
        self,
        bot: Bot,
        method: ConnectMethod[ConnectType],
        type_request: str,
        path: str,
        timeout: Optional[int] = None,
    ) -> ConnectType:  # pragma: no cover
        """
        Make request to Connect Bot API

        :param bot: Bot instance
        :param method: Method instance
        :param type_request: Тип запроса
        :param path: URL
        :param timeout: Request timeout
        :return:
        :raise ConnectApiError:
        """
        pass

    @abc.abstractmethod
    async def stream_content(
        self,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        timeout: int = 30,
        chunk_size: int = 65536,
        raise_for_status: bool = True,
    ) -> AsyncGenerator[bytes, None]:  # pragma: no cover
        """
        Stream reader
        """
        yield b""

    def prepare_value(
        self,
        value: Any,
        bot: Bot,
        files: Dict[str, Any],
        _dumps_json: bool = True,
    ) -> Any:
        """
        Prepare value before send
        """
        if value is None:
            return None
        if isinstance(value, str):
            return value
        if isinstance(value, InputFile):
            key = secrets.token_urlsafe(10)
            files[key] = value
            return f"attach://{key}"
        if isinstance(value, dict):
            value = {
                key: prepared_item
                for key, item in value.items()
                if (
                    prepared_item := self.prepare_value(
                        item, bot=bot, files=files, _dumps_json=False
                    )
                )
                is not None
            }
            if _dumps_json:
                return self.json_dumps(value)
            return value
        if isinstance(value, list):
            value = [
                prepared_item
                for item in value
                if (
                    prepared_item := self.prepare_value(
                        item, bot=bot, files=files, _dumps_json=False
                    )
                )
                is not None
            ]
            if _dumps_json:
                return self.json_dumps(value)
            return value
        if isinstance(value, datetime.timedelta):
            now = datetime.datetime.now()
            return str(round((now + value).timestamp()))
        if isinstance(value, datetime.datetime):
            return str(round(value.timestamp()))
        if isinstance(value, Enum):
            return self.prepare_value(value.value, bot=bot, files=files)

        if _dumps_json:
            return self.json_dumps(value)
        return value

    async def __call__(
        self,
        bot: Bot,
        method: ConnectMethod[ConnectType],
        type_request: str,
        path: str,
        timeout: Optional[int] = None
    ) -> ConnectType:
        middleware = self.middleware.wrap_middlewares(self.make_request, timeout=timeout,
                                                      type_request=type_request, path=path)
        return cast(ConnectType, await middleware(bot, method))

    async def __aenter__(self) -> BaseSession:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()



import asyncio as _asyncio
from contextlib import suppress

from . import enums, methods, types
from .client import session
from .client.bot import Bot
from .dispatcher.dispatcher import Dispatcher
from .dispatcher.middlewares.base import BaseMiddleware
from .dispatcher.router import Router
from .utils.magic_filter import MagicFilter

with suppress(ImportError):
    import uvloop as _uvloop

    _asyncio.set_event_loop_policy(_uvloop.EventLoopPolicy())


F = MagicFilter()

__all__ = (
    "types",
    "methods",
    "enums",
    "Bot",
    "session",
    "Dispatcher",
    "Router",
    "BaseMiddleware",
    "F",
)



from __future__ import annotations

import asyncio
import json
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from aiohttp import BasicAuth, ClientError, ClientSession, FormData, TCPConnector
from aiohttp.hdrs import USER_AGENT
from aiohttp.http import SERVER_SOFTWARE

from .base import BaseSession
from ...methods import ConnectMethod
from ...methods.base import ConnectType
from ...types import InputFile
from ...exceptions import ConnectNetworkError

if TYPE_CHECKING:
    from ..bot import Bot

_ProxyBasic = Union[str, Tuple[str, BasicAuth]]
_ProxyChain = Iterable[_ProxyBasic]
_ProxyType = Union[_ProxyChain, _ProxyBasic]


def _retrieve_basic(basic: _ProxyBasic) -> Dict[str, Any]:
    from aiohttp_socks.utils import parse_proxy_url  # type: ignore

    proxy_auth: Optional[BasicAuth] = None

    if isinstance(basic, str):
        proxy_url = basic
    else:
        proxy_url, proxy_auth = basic

    proxy_type, host, port, username, password = parse_proxy_url(proxy_url)
    if isinstance(proxy_auth, BasicAuth):
        username = proxy_auth.login
        password = proxy_auth.password

    return {
        "proxy_type": proxy_type,
        "host": host,
        "port": port,
        "username": username,
        "password": password,
        "rdns": True,
    }


def _prepare_connector(chain_or_plain: _ProxyType) -> Tuple[Type["TCPConnector"], Dict[str, Any]]:
    from aiohttp_socks import ChainProxyConnector, ProxyConnector, ProxyInfo     # type: ignore

    # since tuple is Iterable(compatible with _ProxyChain) object, we assume that
    # user wants chained proxies if tuple is a pair of string(url) and BasicAuth
    if isinstance(chain_or_plain, str) or (
        isinstance(chain_or_plain, tuple) and len(chain_or_plain) == 2
    ):
        chain_or_plain = cast(_ProxyBasic, chain_or_plain)
        return ProxyConnector, _retrieve_basic(chain_or_plain)

    chain_or_plain = cast(_ProxyChain, chain_or_plain)
    infos: List[ProxyInfo] = []
    for basic in chain_or_plain:
        infos.append(ProxyInfo(**_retrieve_basic(basic)))

    return ChainProxyConnector, {"proxy_infos": infos}


class AiohttpSession(BaseSession):
    def __init__(self, proxy: Optional[_ProxyType] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self._session: Optional[ClientSession] = None
        self._connector_type: Type[TCPConnector] = TCPConnector
        self._should_reset_connector = True  # flag determines connector state
        self._proxy: Optional[_ProxyType] = None

        if proxy is not None:
            try:
                self._setup_proxy_connector(proxy)
            except ImportError as exc:  # pragma: no cover
                raise RuntimeError(
                    "In order to use aiohttp client for proxy requests, install "
                    "https://pypi.org/project/aiohttp-socks/"
                ) from exc

    def _setup_proxy_connector(self, proxy: _ProxyType) -> None:
        self._proxy = proxy

    @property
    def proxy(self) -> Optional[_ProxyType]:
        return self._proxy

    @proxy.setter
    def proxy(self, proxy: _ProxyType) -> None:
        self._setup_proxy_connector(proxy)
        self._should_reset_connector = True

    async def create_session(self) -> ClientSession:
        if self._should_reset_connector:
            await self.close()

        if self._session is None or self._session.closed:
            self._session = ClientSession(
                headers={
                    USER_AGENT: f"{SERVER_SOFTWARE}",
                },
            )
            self._should_reset_connector = False

        return self._session

    async def close(self) -> None:
        if self._session is not None and not self._session.closed:
            await self._session.close()

            # Wait 250 ms for the underlying SSL connections to close
            # https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
            await asyncio.sleep(0.25)

    def build_form_data(self, bot: Bot, method: ConnectMethod[ConnectType]) -> FormData:
        form = FormData(quote_fields=False)
        files: Dict[str, InputFile] = {}

        new_json = {}
        for key, value in method.model_dump(warnings=False, exclude_none=True).items():
            value = self.prepare_value(value, bot=bot, files=files)
            new_json.update({key: value})
        form.add_field('meta', json.dumps(new_json), content_type='application/json')

        for key, value in files.items():
            form.add_field("file", value.read(bot), filename=value.filename or key)
        return form

    async def make_request(
        self, bot: Bot, method: ConnectMethod[ConnectType], type_request: str, path: str, timeout: Optional[int] = None
    ) -> ConnectType:
        session = await self.create_session()

        url = bot.base + path
        form = self.build_form_data(bot=bot, method=method)
        json_data = method.model_dump()

        try:
            if type_request == "POST-With-Attach":
                async with session.post(
                    auth=bot.auth, url=url, data=form,
                        timeout=self.timeout if timeout is None else timeout
                ) as resp:
                    raw_result = await resp.text()
            if type_request == "POST":
                async with session.post(
                    auth=bot.auth, url=url, json=json_data,
                        timeout=self.timeout if timeout is None else timeout
                ) as resp:
                    raw_result = await resp.text()
            elif type_request == "DELETE":
                async with session.delete(
                    auth=bot.auth, url=url, params=method.model_dump(exclude_none=True),
                        timeout=self.timeout if timeout is None else timeout
                ) as resp:
                    raw_result = await resp.text()
            elif type_request == "GET":
                async with session.get(
                    auth=bot.auth, url=url, params=method.model_dump(exclude_none=True),
                        timeout=self.timeout if timeout is None else timeout
                ) as resp:
                    raw_result = await resp.text()
            elif type_request == "PUT":
                async with session.put(
                    auth=bot.auth, url=url, json=json_data,
                        timeout=self.timeout if timeout is None else timeout
                ) as resp:
                    raw_result = await resp.text()
        except asyncio.TimeoutError:
            raise ConnectNetworkError(method=method, message="Request timeout error")
        except ClientError as e:
            raise ConnectNetworkError(method=method, message=f"{type(e).__name__}: {e}")
        response = self.check_response(
            bot=bot, method=method, status_code=resp.status, content=raw_result
        )
        return cast(ConnectType, response.result)

    async def stream_content(
        self,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        timeout: int = 30,
        chunk_size: int = 65536,
        raise_for_status: bool = True,
    ) -> AsyncGenerator[bytes, None]:
        if headers is None:
            headers = {}

        session = await self.create_session()

        async with session.get(
            url, timeout=timeout, headers=headers, raise_for_status=raise_for_status
        ) as resp:
            async for chunk in resp.content.iter_chunked(chunk_size):
                yield chunk

    async def __aenter__(self) -> AiohttpSession:
        await self.create_session()
        return self
